#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""FastAPI application models."""

from datetime import datetime
from typing import Any, Dict, Optional
from uuid import UUID

from pydantic import BaseModel, Field

from zenml.logger import get_logger

logger = get_logger(__name__)


class DeploymentInvocationResponseMetadata(BaseModel):
    """Pipeline invoke response metadata model."""

    deployment_id: UUID = Field(title="The ID of the deployment.")
    deployment_name: str = Field(title="The name of the deployment.")
    snapshot_id: UUID = Field(title="The ID of the snapshot.")
    snapshot_name: Optional[str] = Field(
        default=None, title="The name of the snapshot."
    )
    pipeline_name: str = Field(title="The name of the pipeline.")
    run_id: Optional[UUID] = Field(
        default=None, title="The ID of the pipeline run."
    )
    run_index: Optional[int] = Field(
        default=None, title="The index of the pipeline run."
    )
    run_name: Optional[str] = Field(
        default=None, title="The name of the pipeline run."
    )
    parameters_used: Dict[str, Any] = Field(
        title="The parameters used for the pipeline execution."
    )


class BaseDeploymentInvocationRequest(BaseModel):
    """Base pipeline invoke request model."""

    parameters: BaseModel = Field(
        title="The parameters for the pipeline execution."
    )
    run_name: Optional[str] = Field(
        default=None, title="Custom name for the pipeline run."
    )
    timeout: int = Field(
        default=300, title="The timeout for the pipeline execution."
    )
    skip_artifact_materialization: bool = Field(
        default=False,
        title="Whether to keep outputs in memory for fast access instead of "
        "storing them as artifacts.",
    )


class BaseDeploymentInvocationResponse(BaseModel):
    """Base pipeline invoke response model."""

    success: bool = Field(
        title="Whether the pipeline execution was successful."
    )
    outputs: Optional[Dict[str, Any]] = Field(
        default=None,
        title="The outputs of the pipeline execution, if the pipeline execution "
        "was successful.",
    )
    execution_time: float = Field(
        title="The time taken to execute the pipeline."
    )
    metadata: DeploymentInvocationResponseMetadata = Field(
        title="The metadata of the pipeline execution."
    )
    error: Optional[str] = Field(
        default=None,
        title="The error that occurred, if the pipeline invocation failed.",
    )


class PipelineInfo(BaseModel):
    """Pipeline info model."""

    name: str = Field(title="The name of the pipeline.")
    parameters: Optional[Dict[str, Any]] = Field(
        default=None, title="The parameters of the pipeline."
    )
    input_schema: Optional[Dict[str, Any]] = Field(
        default=None, title="The input schema of the pipeline."
    )
    output_schema: Optional[Dict[str, Any]] = Field(
        default=None, title="The output schema of the pipeline."
    )


class DeploymentInfo(BaseModel):
    """Deployment info model."""

    id: UUID = Field(title="The ID of the deployment.")
    name: str = Field(title="The name of the deployment.")
    auth_enabled: bool = Field(
        title="Whether the deployment is authenticated."
    )


class SnapshotInfo(BaseModel):
    """Snapshot info model."""

    id: UUID = Field(title="The ID of the snapshot.")
    name: Optional[str] = Field(
        default=None, title="The name of the snapshot."
    )


class AppInfo(BaseModel):
    """App info model."""

    app_runner_flavor: str
    docs_url_path: str
    redoc_url_path: str
    invoke_url_path: str
    health_url_path: str
    info_url_path: str
    metrics_url_path: str


class ServiceInfo(BaseModel):
    """Service info model."""

    deployment: DeploymentInfo = Field(
        title="The deployment of the pipeline service."
    )
    snapshot: SnapshotInfo = Field(
        title="The snapshot of the pipeline service."
    )
    pipeline: PipelineInfo = Field(
        title="The pipeline of the pipeline service."
    )
    app: AppInfo = Field(title="The deployment application")
    total_executions: int = Field(
        title="The total number of pipeline executions."
    )
    last_execution_time: Optional[datetime] = Field(
        default=None, title="The time of the last pipeline execution."
    )
    status: str = Field(title="The status of the pipeline service.")
    uptime: float = Field(title="The uptime of the pipeline service.")


class ExecutionMetrics(BaseModel):
    """Execution metrics model."""

    total_executions: int = Field(
        title="The total number of pipeline executions."
    )
    last_execution_time: Optional[datetime] = Field(
        default=None, title="The time of the last pipeline execution."
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes deployer implementation."""

import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from kubernetes import client as k8s_client
from kubernetes import config as k8s_config
from kubernetes.client.rest import ApiException
from pydantic import BaseModel, ConfigDict

from zenml.deployers.containerized_deployer import ContainerizedDeployer
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentDeprovisionError,
    DeploymentLogsNotFoundError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.deployers.server.entrypoint_configuration import (
    DEPLOYMENT_ID_OPTION,
    DeploymentEntrypointConfiguration,
)
from zenml.entrypoints.base_entrypoint_configuration import (
    SNAPSHOT_ID_OPTION,
)
from zenml.enums import DeploymentStatus, StackComponentType
from zenml.integrations.kubernetes import kube_utils
from zenml.integrations.kubernetes.flavors.kubernetes_deployer_flavor import (
    KubernetesDeployerConfig,
    KubernetesDeployerSettings,
)
from zenml.integrations.kubernetes.k8s_applier import (
    KubernetesApplier,
    ProvisioningError,
    ResourceInventoryItem,
)
from zenml.integrations.kubernetes.manifest_utils import (
    build_namespace_manifest,
    build_secret_manifest,
)
from zenml.integrations.kubernetes.pod_settings import (
    KubernetesPodSettings,
)
from zenml.integrations.kubernetes.serialization_utils import (
    normalize_resource_to_dict,
)
from zenml.integrations.kubernetes.template_engine import (
    KubernetesTemplateEngine,
)
from zenml.logger import get_logger
from zenml.models import (
    DeploymentOperationalState,
    DeploymentResponse,
    PipelineSnapshotResponse,
)
from zenml.stack import StackValidator

if TYPE_CHECKING:
    from zenml.stack import Stack

logger = get_logger(__name__)

MAX_LOAD_BALANCER_TIMEOUT = 600  # 10 minutes


class _DeploymentCtx(BaseModel):
    """Deployment context."""

    settings: KubernetesDeployerSettings
    snapshot: PipelineSnapshotResponse
    namespace: str
    resource_name: str
    deployment_id: str
    labels: Dict[str, str]
    secret_name: str
    image: str

    model_config = ConfigDict(frozen=True)


class KubernetesDeployer(ContainerizedDeployer):
    """Kubernetes deployer using template-based resource management."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the Kubernetes deployer.

        Args:
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self._k8s_client: Optional[k8s_client.ApiClient] = None
        self._applier: Optional[KubernetesApplier] = None

    # ========================================================================
    # Kubernetes Client Management
    # ========================================================================

    def get_kube_client(
        self, incluster: Optional[bool] = None
    ) -> k8s_client.ApiClient:
        """Getter for the Kubernetes API client.

        Args:
            incluster: Whether to use the in-cluster config or not. Overrides
                the `incluster` setting in the config.

        Returns:
            The Kubernetes API client.

        Raises:
            RuntimeError: if the Kubernetes connector behaves unexpectedly.
        """
        if incluster is None:
            incluster = self.config.incluster

        if incluster:
            try:
                kube_utils.load_kube_config(
                    incluster=incluster,
                )
                self._k8s_client = k8s_client.ApiClient()
                return self._k8s_client
            except Exception as e:
                if self.connector:
                    message = (
                        "Falling back to using the linked service connector "
                        "configuration."
                    )
                elif self.config.kubernetes_context:
                    message = (
                        f"Falling back to using the configured "
                        f"'{self.config.kubernetes_context}' kubernetes context."
                    )
                else:
                    raise RuntimeError(
                        f"The orchestrator failed to load the in-cluster "
                        f"Kubernetes configuration and there is no service "
                        f"connector or kubernetes_context to fall back to: {e}"
                    ) from e

                logger.debug(
                    f"Could not load the in-cluster Kubernetes configuration: "
                    f"{e}. {message}"
                )

        if self._k8s_client and not self.connector_has_expired():
            return self._k8s_client

        connector = self.get_connector()
        if connector:
            client = connector.connect()
            if not isinstance(client, k8s_client.ApiClient):
                raise RuntimeError(
                    f"Expected a k8s_client.ApiClient while trying to use the "
                    f"linked connector, but got {type(client)}."
                )
            self._k8s_client = client
        else:
            kube_utils.load_kube_config(
                context=self.config.kubernetes_context,
            )
            self._k8s_client = k8s_client.ApiClient()

        return self._k8s_client

    @property
    def k8s_core_api(self) -> k8s_client.CoreV1Api:
        """Get Kubernetes Core V1 API client.

        Returns:
            Kubernetes Core V1 API client.
        """
        return k8s_client.CoreV1Api(self.get_kube_client())

    @property
    def k8s_applier(self) -> KubernetesApplier:
        """Get or create Kubernetes Applier instance.

        Returns:
            Kubernetes Applier instance.
        """
        if self.connector_has_expired():
            self._applier = None

        if not self._applier:
            self._applier = KubernetesApplier(
                api_client=self.get_kube_client()
            )
        return self._applier

    # ========================================================================
    # Configuration and Validation
    # ========================================================================

    @property
    def config(self) -> KubernetesDeployerConfig:
        """Get the deployer configuration.

        Returns:
            The deployer configuration.
        """
        return cast(KubernetesDeployerConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type[KubernetesDeployerSettings]]:
        """Return the settings class.

        Returns:
            The settings class for this deployer.
        """
        return KubernetesDeployerSettings

    def get_kubernetes_contexts(self) -> Tuple[List[str], str]:
        """Get list of configured Kubernetes contexts and the active context.

        Raises:
            RuntimeError: if the Kubernetes configuration cannot be loaded.

        Returns:
            context_name: List of configured Kubernetes contexts
            active_context_name: Name of the active Kubernetes context.
        """
        try:
            contexts, active_context = k8s_config.list_kube_config_contexts()
        except k8s_config.config_exception.ConfigException as e:
            raise RuntimeError(
                "Could not load the Kubernetes configuration"
            ) from e

        context_names = [c["name"] for c in contexts]
        active_context_name = active_context["name"]
        return context_names, active_context_name

    @property
    def validator(self) -> Optional[StackValidator]:
        """Stack validator for the deployer.

        Returns:
            Stack validator instance.
        """

        def _validate(stack: "Stack") -> Tuple[bool, str]:
            container_registry = stack.container_registry
            if not container_registry:
                return False, "Container registry is required"

            kubernetes_context = self.config.kubernetes_context
            msg = f"'{self.name}' Kubernetes deployer error: "

            if not self.connector:
                if self.config.incluster:
                    # No service connector or kubernetes_context is needed when
                    # the deployer is being used from within a Kubernetes
                    # cluster.
                    pass
                elif kubernetes_context:
                    contexts, active_context = self.get_kubernetes_contexts()

                    if kubernetes_context not in contexts:
                        return False, (
                            f"{msg}could not find a Kubernetes context named "
                            f"'{kubernetes_context}' in the local "
                            "Kubernetes configuration. Please make sure that "
                            "the Kubernetes cluster is running and that the "
                            "kubeconfig file is configured correctly. To list "
                            "all configured contexts, run:\n\n"
                            "  `kubectl config get-contexts`\n"
                        )
                    if kubernetes_context != active_context:
                        warning_msg = "".join(
                            [
                                f"{msg}the Kubernetes context '{kubernetes_context}' configured for the ",
                                "Kubernetes deployer is not the same as the active context in the local ",
                                "Kubernetes configuration. If this is not deliberate, you should update the ",
                                "deployer's `kubernetes_context` field by running:\n\n",
                                f"  `zenml deployer update {self.name} --kubernetes_context={active_context}`\n",
                                "To list all configured contexts, run:\n\n",
                                "  `kubectl config get-contexts`\n",
                                "To set the active context to be the same as the one configured in the ",
                                "Kubernetes deployer and silence this warning, run:\n\n",
                                f"  `kubectl config use-context {kubernetes_context}`\n",
                            ]
                        )
                        logger.warning(warning_msg)
                else:
                    return False, (
                        f"{msg}you must either link this deployer to a "
                        "Kubernetes service connector (see the 'zenml "
                        "deployer connect' CLI command), explicitly set "
                        "the `kubernetes_context` attribute to the name of the "
                        "Kubernetes config context pointing to the cluster "
                        "where you would like to deploy, or set the "
                        "`incluster` attribute to `True`."
                    )

            silence_local_validations_msg = (
                f"To silence this warning, set the "
                f"`skip_local_validations` attribute to True in the "
                f"deployer configuration by running:\n\n"
                f"  'zenml deployer update {self.name} "
                f"--skip_local_validations=True'\n"
            )

            if (
                not self.config.skip_local_validations
                and not self.config.is_local
            ):
                # If the deployer is not running in a local k3d cluster,
                # we cannot have any other local components in our stack,
                # because we cannot mount the local path into the container.
                # This may result in problems when running deployments, because
                # the local components will not be available inside the
                # kubernetes containers.

                # Go through all stack components and identify those that
                # advertise a local path where they persist information that
                # they need to be available when running deployments.
                for stack_comp in stack.components.values():
                    if stack_comp.local_path is None:
                        continue
                    return False, (
                        f"{msg}the Kubernetes deployer is configured to "
                        f"run deployments in a remote Kubernetes cluster but the "
                        f"'{stack_comp.name}' {stack_comp.type.value} "
                        f"is a local stack component "
                        f"and will not be available in the Kubernetes deployment "
                        f"containers.\nPlease ensure that you always use non-local "
                        f"stack components with a remote Kubernetes deployer, "
                        f"otherwise you may run into deployment execution "
                        f"problems. You should use a flavor of "
                        f"{stack_comp.type.value} other than "
                        f"'{stack_comp.flavor}'.\n"
                        + silence_local_validations_msg
                    )

                # If the deployer is remote, the container registry must
                # also be remote.
                if container_registry.config.is_local:
                    return False, (
                        f"{msg}the Kubernetes deployer is configured to "
                        "run deployments in a remote Kubernetes cluster but the "
                        f"'{container_registry.name}' container registry URI "
                        f"'{container_registry.config.uri}' "
                        f"points to a local container registry. Please ensure "
                        f"that you always use non-local stack components with "
                        f"a remote Kubernetes deployer, otherwise you will "
                        f"run into problems. You should use a flavor of "
                        f"container registry other than "
                        f"'{container_registry.flavor}'.\n"
                        + silence_local_validations_msg
                    )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.IMAGE_BUILDER,
                StackComponentType.CONTAINER_REGISTRY,
            },
            custom_validation_function=_validate,
        )

    # ========================================================================
    # Template Context Building
    # ========================================================================

    def _build_template_context(
        self,
        deployment: DeploymentResponse,
        settings: KubernetesDeployerSettings,
        resource_name: str,
        namespace: str,
        labels: Dict[str, str],
        image: str,
        env_vars: Dict[str, str],
        secret_env_vars: Dict[str, str],
        secret_name: str,
        resource_requests: Dict[str, str],
        resource_limits: Dict[str, str],
        replicas: int,
    ) -> Dict[str, Any]:
        """Build template rendering context from deployment configuration.

        Args:
            deployment: The deployment response object.
            settings: Kubernetes deployer settings.
            resource_name: Sanitized resource name.
            namespace: Kubernetes namespace.
            labels: Resource labels.
            image: Container image.
            env_vars: Environment variables dict.
            secret_env_vars: Secret environment variables dict.
            secret_name: Kubernetes secret name.
            resource_requests: Resource requests dict.
            resource_limits: Resource limits dict.
            replicas: Number of replicas.

        Returns:
            Template context dictionary.
        """
        resources = {}
        if resource_requests:
            resources["requests"] = resource_requests
        if resource_limits:
            resources["limits"] = resource_limits

        secret_env_list = []
        for key in secret_env_vars.keys():
            secret_env_list.append(
                {
                    "name": key,
                    "valueFrom": {
                        "secretKeyRef": {
                            "name": secret_name,
                            "key": key,
                        }
                    },
                }
            )

        pod_settings_for_template = settings.pod_settings
        if secret_env_list and settings.pod_settings:
            pod_settings_for_template = settings.pod_settings.model_copy(
                update={
                    "env": (settings.pod_settings.env or []) + secret_env_list
                }
            )
        elif secret_env_list:
            pod_settings_for_template = KubernetesPodSettings(
                env=secret_env_list
            )

        command = settings.command
        if not command:
            command = cast(
                Any,
                DeploymentEntrypointConfiguration.get_entrypoint_command(),
            )

        args = settings.args
        if not args:
            entrypoint_kwargs = {DEPLOYMENT_ID_OPTION: deployment.id}
            if deployment.snapshot:
                entrypoint_kwargs[SNAPSHOT_ID_OPTION] = deployment.snapshot.id
            args = cast(
                Any,
                DeploymentEntrypointConfiguration.get_entrypoint_arguments(
                    **entrypoint_kwargs
                ),
            )

        context = {
            "deployment": deployment,
            "settings": settings,
            "name": resource_name,
            "namespace": namespace,
            "labels": labels,
            "image": image,
            "command": command,
            "args": args,
            "env": env_vars,
            "resources": resources if resources else None,
            "pod_settings": pod_settings_for_template,
            "replicas": replicas,
        }

        return context

    def _get_template_path(
        self, builtin_name: str, custom_path: Optional[str]
    ) -> str:
        """Get template path (custom or built-in).

        Args:
            builtin_name: Name of built-in template file.
            custom_path: Optional custom template path.

        Returns:
            Path to the template file.
        """
        if custom_path:
            return custom_path

        from pathlib import Path

        builtin_dir = Path(__file__).parent.parent / "templates" / "kubernetes"
        return str(builtin_dir / builtin_name)

    # ========================================================================
    # Secret Key Validation
    # ========================================================================

    def _validate_secret_key(
        self,
        key: str,
        secret_key_map: Dict[str, str],
        existing_env_vars: Optional[Dict[str, str]] = None,
    ) -> str:
        """Validate secret key is a valid K8s env var name and check for collisions.

        Args:
            key: Original secret key name.
            secret_key_map: Dictionary tracking validated key mappings to detect collisions.
            existing_env_vars: Optional dictionary of existing environment variables
                to check for collisions.

        Returns:
            The validated secret key name.

        Raises:
            DeploymentProvisionError: If key is invalid or causes a collision.
        """
        if not re.match(r"^[A-Za-z_][A-Za-z0-9_]*$", key):
            raise DeploymentProvisionError(
                f"Secret key '{key}' is not a valid Kubernetes environment variable name. "
                f"It must start with a letter or underscore and contain only letters, "
                f"numbers, and underscores. Please rename the secret."
            )

        if key in secret_key_map and secret_key_map[key] != key:
            raise DeploymentProvisionError(
                f"Secret key '{key}' already exists as another secret. "
                f"Please use unique secret names."
            )

        if existing_env_vars and key in existing_env_vars:
            raise DeploymentProvisionError(
                f"Secret key '{key}' conflicts with an existing environment variable. "
                f"Please rename the secret to avoid collision."
            )

        secret_key_map[key] = key
        return key

    # ========================================================================
    # Provisioning
    # ========================================================================

    def _find_resources_to_delete(
        self,
        old_inventory: List[ResourceInventoryItem],
        new_inventory: List[ResourceInventoryItem],
    ) -> List[ResourceInventoryItem]:
        """Find resources in old inventory that are not in new inventory.

        Args:
            old_inventory: Resources from previous deployment.
            new_inventory: Resources from current deployment.

        Returns:
            List of resources that should be deleted.
        """
        new_resource_ids = {
            (item.kind, item.api_version, item.name, item.namespace)
            for item in new_inventory
        }

        resources_to_delete = []
        for item in old_inventory:
            resource_id = (
                item.kind,
                item.api_version,
                item.name,
                item.namespace,
            )
            if resource_id not in new_resource_ids:
                resources_to_delete.append(item)

        return resources_to_delete

    def _wait_for_deployment_readiness(
        self,
        inventory: List[ResourceInventoryItem],
        settings: KubernetesDeployerSettings,
        timeout: int,
    ) -> None:
        """Wait for all deployments and load balancer services to become ready.

        Uses the resource inventory to find all Deployment and LoadBalancer Service
        resources, then waits for each one to become ready.

        Args:
            inventory: List of provisioned resources to wait for.
            settings: Kubernetes deployer settings.
            timeout: Timeout in seconds.

        Raises:
            DeploymentProvisionError: If any deployment doesn't become ready in time.
        """
        if timeout <= 0:
            return

        deployments = [
            item
            for item in inventory
            if item.kind == "Deployment" and item.api_version == "apps/v1"
        ]

        for deployment_item in deployments:
            try:
                logger.info(
                    f"Waiting for Deployment '{deployment_item.name}' "
                    f"in namespace '{deployment_item.namespace}' to become ready "
                    f"(timeout: {timeout}s)"
                )
                if not deployment_item.namespace:
                    raise RuntimeError(
                        f"Deployment '{deployment_item.name}' has no namespace"
                    )
                self.k8s_applier.wait_for_deployment_ready(
                    name=deployment_item.name,
                    namespace=deployment_item.namespace,
                    timeout=timeout,
                    check_interval=settings.deployment_ready_check_interval,
                )
                logger.info(f"Deployment '{deployment_item.name}' is ready")
            except RuntimeError as e:
                raise DeploymentProvisionError(
                    f"Deployment '{deployment_item.name}' in namespace "
                    f"'{deployment_item.namespace}' did not become ready: {e}"
                ) from e

        if settings.wait_for_load_balancer_timeout > 0:
            services = [
                item
                for item in inventory
                if item.kind == "Service" and item.api_version == "v1"
            ]

            lb_timeout = min(
                timeout,
                settings.wait_for_load_balancer_timeout,
                MAX_LOAD_BALANCER_TIMEOUT,
            )

            for service_item in services:
                try:
                    service = self.k8s_applier.get_resource(
                        name=service_item.name,
                        namespace=service_item.namespace,
                        kind="Service",
                        api_version="v1",
                    )
                    if not service:
                        continue

                    service_type = (
                        normalize_resource_to_dict(service)
                        .get("spec", {})
                        .get("type")
                    )
                    if service_type != "LoadBalancer":
                        continue

                    logger.info(
                        f"Waiting for LoadBalancer Service '{service_item.name}' "
                        f"in namespace '{service_item.namespace}' to get external IP "
                        f"(timeout: {lb_timeout}s)"
                    )
                    if not service_item.namespace:
                        raise RuntimeError(
                            f"Service '{service_item.name}' has no namespace"
                        )
                    self.k8s_applier.wait_for_service_loadbalancer_ip(
                        name=service_item.name,
                        namespace=service_item.namespace,
                        timeout=lb_timeout,
                        check_interval=settings.deployment_ready_check_interval,
                    )
                    logger.info(
                        f"LoadBalancer Service '{service_item.name}' has external IP"
                    )
                except RuntimeError:
                    logger.warning(
                        f"LoadBalancer Service '{service_item.name}' did not get "
                        f"external IP within {lb_timeout}s.\n"
                        f"  This can happen if:\n"
                        f"  • You're on a local cluster (minikube, kind, k3d, docker-desktop) that doesn't support LoadBalancers\n"
                        f"  • Your cloud provider is slow to provision the load balancer\n"
                        f"  • There are cloud provider issues (quotas, permissions, etc.)\n"
                        f"\n"
                        f"  Next steps:\n"
                        f"  1. Check if IP appears: kubectl get svc {service_item.name} -n {service_item.namespace}\n"
                        f"  2. Refresh deployment status later to check if IP was assigned: zenml deployment refresh <deployment-name>\n"
                        f"  3. For local clusters, consider using service_type='NodePort' instead\n"
                        f"  4. For immediate access: kubectl port-forward -n {service_item.namespace} service/{service_item.name} <local-port>:<service-port>"
                    )
                except Exception as e:
                    logger.warning(
                        f"Error checking service '{service_item.name}': {e}"
                    )

    def _initialize_deployment_context(
        self, deployment: DeploymentResponse
    ) -> _DeploymentCtx:
        """Initialize basic deployment context from deployment snapshot.

        Sets up instance variables: _settings, _namespace, _resource_name,
        _labels, and _engine (if needed).

        Args:
            deployment: The deployment to initialize context for.

        Raises:
            DeployerError: If deployment has no snapshot.

        Returns:
            The deployment context.
        """
        snapshot = deployment.snapshot
        if not snapshot:
            raise DeployerError(
                f"Deployment '{deployment.name}' has no snapshot"
            )

        settings = cast(
            KubernetesDeployerSettings, self.get_settings(snapshot)
        )
        deployment_id = str(deployment.id)
        resource_name = kube_utils.sanitize_label(f"zenml-{deployment_id}")
        secret_name = resource_name

        labels = {
            "zenml-deployment-id": deployment_id,
            "zenml-deployment-name": kube_utils.sanitize_label(
                deployment.name
            ),
            "managed-by": "zenml",
        }
        if settings.labels:
            labels.update(**settings.labels)

        image = self.get_image(snapshot)

        return _DeploymentCtx(
            settings=settings,
            snapshot=snapshot,
            namespace=settings.namespace,
            deployment_id=deployment_id,
            resource_name=resource_name,
            secret_name=secret_name,
            labels=labels,
            image=image,
        )

    def _prepare_deployment_resources(
        self,
        deployment: DeploymentResponse,
        environment: Dict[str, str],
        secrets: Dict[str, str],
    ) -> Tuple[_DeploymentCtx, List[Dict[str, Any]]]:
        """Prepare all resources needed for deployment.

        Args:
            deployment: The deployment to prepare.
            environment: Environment variables.
            secrets: Secret environment variables.

        Returns:
            List of rendered kubernetes resources.

        Raises:
            DeploymentProvisionError: If preparation fails.
        """
        ctx = self._initialize_deployment_context(deployment)
        engine = KubernetesTemplateEngine()

        try:
            resource_requests, resource_limits, replicas = (
                kube_utils.convert_resource_settings_to_k8s_format(
                    ctx.snapshot.pipeline_configuration.resource_settings
                )
            )
        except ValueError as e:
            raise DeploymentProvisionError(
                f"Invalid resource settings for deployment '{deployment.name}': {e}"
            ) from e

        secret_key_map: Dict[str, str] = {}
        validated_secrets = {}
        for key, value in secrets.items():
            validated_key = self._validate_secret_key(
                key, secret_key_map, environment
            )
            validated_secrets[validated_key] = value

        context = self._build_template_context(
            deployment=deployment,
            settings=ctx.settings,
            resource_name=ctx.resource_name,
            namespace=ctx.namespace,
            labels=ctx.labels,
            image=ctx.image,
            env_vars=environment,
            secret_env_vars=validated_secrets,
            secret_name=ctx.secret_name,
            resource_requests=resource_requests,
            resource_limits=resource_limits,
            replicas=replicas,
        )

        rendered_resources: List[Dict[str, Any]] = [
            build_namespace_manifest(
                namespace=ctx.namespace, labels=ctx.labels
            )
        ]

        if validated_secrets:
            secret_manifest = build_secret_manifest(
                name=ctx.secret_name,
                data=validated_secrets,
                namespace=ctx.namespace,
            )
            rendered_resources.append(
                self.k8s_applier.api_client.sanitize_for_serialization(
                    secret_manifest
                )
            )

        deployment_template = self._get_template_path(
            "deployment.yaml.j2", ctx.settings.custom_deployment_template_file
        )
        rendered_resources.extend(
            engine.render_template(deployment_template, context)
        )

        service_template = self._get_template_path(
            "service.yaml.j2", ctx.settings.custom_service_template_file
        )
        rendered_resources.extend(
            engine.render_template(service_template, context)
        )

        for additional_resource in ctx.settings.additional_resources or []:
            try:
                additional_docs = engine.render_template(
                    additional_resource, context
                )
                if additional_docs:
                    rendered_resources.extend(additional_docs)
                else:
                    logger.debug(
                        f"Additional resource '{additional_resource}' rendered "
                        f"to no documents (empty or all conditionals False)"
                    )
            except Exception as e:
                error_msg = f"Failed to render additional resource '{additional_resource}': {e}"
                if ctx.settings.strict_additional_resources:
                    raise DeploymentProvisionError(error_msg) from e
                else:
                    logger.warning(
                        f"{error_msg}. Skipping (strict_additional_resources=False)"
                    )

        return ctx, rendered_resources

    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Provision a Kubernetes deployment.

        Args:
            deployment: The deployment to provision.
            stack: The stack the pipeline will be deployed on.
            environment: Environment variables.
            secrets: Secret environment variables.
            timeout: Timeout in seconds.

        Returns:
            The operational state of the deployment.

        Raises:
            DeploymentProvisionError: If provisioning fails.
        """
        ctx, rendered_resources = self._prepare_deployment_resources(
            deployment, environment, secrets
        )

        stored_metadata = deployment.deployment_metadata or {}
        old_inventory_data = stored_metadata.get("resource_inventory")

        try:
            created_objects, new_inventory = self.k8s_applier.provision(
                rendered_resources,
                default_namespace=ctx.namespace,
                timeout=timeout,
            )
            logger.info(
                f"Successfully created/updated {len(created_objects)} "
                f"Kubernetes resource(s) for deployment '{deployment.name}'"
            )

            if old_inventory_data:
                old_inventory = [
                    ResourceInventoryItem(**item)
                    for item in old_inventory_data
                ]
                resources_to_delete = self._find_resources_to_delete(
                    old_inventory, new_inventory
                )
                if resources_to_delete:
                    logger.info(
                        f"Cleaning up {len(resources_to_delete)} obsolete "
                        f"resource(s) from previous deployment of "
                        f"'{deployment.name}'"
                    )
                    delete_result = self.k8s_applier.delete_from_inventory(
                        inventory=resources_to_delete,
                        propagation_policy="Foreground",
                    )
                    logger.info(
                        "Obsolete resource cleanup result: "
                        f"deleted={delete_result.deleted_count} "
                        f"skipped={delete_result.skipped_count} "
                        f"failed={delete_result.failed_count}"
                    )

            self._wait_for_deployment_readiness(
                inventory=new_inventory,
                settings=ctx.settings,
                timeout=timeout,
            )

            state = self._get_deployment_state_from_inventory(
                inventory=new_inventory,
                settings=ctx.settings,
            )

            if state.metadata is None:
                state.metadata = {}
            state.metadata["resource_inventory"] = [
                item.model_dump() for item in new_inventory
            ]

            return state

        except ProvisioningError as e:
            new_inventory = e.inventory
            for error_line in e.errors:
                logger.error(f"  • {error_line}")

            if ctx.settings.atomic_provision and new_inventory:
                logger.warning(
                    f"Atomic provision enabled – rolling back "
                    f"{len(new_inventory)} resource(s) for '{deployment.name}'"
                )
                delete_result = self.k8s_applier.delete_from_inventory(
                    inventory=new_inventory,
                    propagation_policy="Foreground",
                )
                raise DeploymentProvisionError(
                    "Kubernetes provisioning failed for deployment "
                    f"'{deployment.name}' with atomic_provision=True.\n"
                    f"Errors ({len(e.errors)}):\n"
                    + "\n".join(f"  • {err}" for err in e.errors)
                    + "\n\nRollback summary:\n"
                    f"  deleted={delete_result.deleted_count}, "
                    f"skipped={delete_result.skipped_count}, "
                    f"failed={delete_result.failed_count}"
                ) from e

            deployment_state = DeploymentOperationalState(
                status=DeploymentStatus.ERROR,
                metadata={
                    "namespace": ctx.namespace,
                    "labels": ctx.labels,
                    "resource_inventory": [
                        item.model_dump() for item in new_inventory
                    ],
                    "errors": e.errors,
                },
            )
            deployment = self._update_deployment(deployment, deployment_state)

            raise DeploymentProvisionError(
                f"Kubernetes provisioning failed for deployment "
                f"'{deployment.name}'. Errors:\n"
                + "\n".join(f"  • {err}" for err in e.errors)
            ) from e

        except Exception as e:
            raise DeploymentProvisionError(
                f"Failed to provision deployment '{deployment.name}': {e}"
            ) from e

    # ========================================================================
    # State Management
    # ========================================================================

    def _get_deployment_state_from_inventory(
        self,
        inventory: List[ResourceInventoryItem],
        settings: KubernetesDeployerSettings,
    ) -> DeploymentOperationalState:
        """Get deployment state from a resource inventory.

        This helper method extracts state determination logic so it can be
        used both during provisioning (when inventory is in memory) and when
        querying (when inventory comes from stored metadata).

        Args:
            inventory: The resource inventory to check.
            settings: The deployer settings.

        Returns:
            The operational state.

        Raises:
            DeploymentNotFoundError: If no deployment resources found in cluster.
            DeployerError: If an error occurs checking resources.
        """
        deployment_items = [
            item
            for item in inventory
            if item.kind == "Deployment" and item.api_version == "apps/v1"
        ]

        service_items = [
            item
            for item in inventory
            if item.kind == "Service" and item.api_version == "v1"
        ]

        try:
            status = DeploymentStatus.PENDING
            all_ready = True
            any_exists = False

            for deployment_item in deployment_items:
                k8s_deployment = self.k8s_applier.get_resource(
                    name=deployment_item.name,
                    namespace=deployment_item.namespace,
                    kind="Deployment",
                    api_version="apps/v1",
                )

                if not k8s_deployment:
                    logger.debug(
                        f"Deployment '{deployment_item.name}' not found in cluster"
                    )
                    all_ready = False
                    continue

                any_exists = True

                deployment_dict = normalize_resource_to_dict(k8s_deployment)
                status_data = deployment_dict.get("status", {})
                spec_data = deployment_dict.get("spec", {})
                available = status_data.get("availableReplicas", 0)
                desired = spec_data.get("replicas", 0)

                logger.debug(
                    f"Deployment '{deployment_item.name}': "
                    f"available={available}, desired={desired}"
                )

                if desired == 0:
                    logger.debug(
                        f"Deployment '{deployment_item.name}' is scaled to zero "
                        f"(idle-ready state)"
                    )
                elif available < desired:
                    all_ready = False

            if not any_exists:
                raise DeploymentNotFoundError(
                    "No deployment resources found in cluster"
                )

            if all_ready:
                status = DeploymentStatus.RUNNING

            url = None
            namespace = settings.namespace

            for service_item in service_items:
                k8s_service = self.k8s_applier.get_resource(
                    name=service_item.name,
                    namespace=service_item.namespace,
                    kind="Service",
                    api_version="v1",
                )

                if k8s_service:
                    service_url = kube_utils.build_service_url(
                        core_api=self.k8s_core_api,
                        service=k8s_service,
                        namespace=service_item.namespace or namespace,
                        ingress=None,
                    )
                    if service_url:
                        url = service_url
                        break

            metadata = {
                "namespace": namespace,
                "labels": settings.labels,
            }

            return DeploymentOperationalState(
                status=status,
                url=url,
                metadata=metadata,
            )

        except ApiException as e:
            if e.status == 404:
                raise DeploymentNotFoundError(
                    "Deployment resources not found in cluster"
                )
            raise DeployerError(f"Failed to get deployment state: {e}")

    def do_get_deployment_state(
        self, deployment: DeploymentResponse
    ) -> DeploymentOperationalState:
        """Get deployment state based on inventory.

        This method checks ALL resources in the inventory to determine status:
        - Checks all Deployment resources to see if they're ready
        - Checks all Service resources to get URLs
        - Works with both built-in and custom templates

        Args:
            deployment: The deployment.

        Returns:
            The operational state.

        Raises:
            DeploymentNotFoundError: If no resources found in cluster.
        """
        ctx = self._initialize_deployment_context(deployment)

        stored_metadata = deployment.deployment_metadata or {}
        inventory_data = stored_metadata.get("resource_inventory")

        if not inventory_data:
            raise DeploymentNotFoundError(
                f"Deployment '{deployment.name}' has no stored resource inventory"
            )

        inventory = [
            ResourceInventoryItem(**item)
            for item in stored_metadata["resource_inventory"]
        ]

        state = self._get_deployment_state_from_inventory(
            inventory=inventory,
            settings=ctx.settings,
        )

        if state.metadata is None:
            state.metadata = {}
        state.metadata["resource_inventory"] = stored_metadata[
            "resource_inventory"
        ]

        return state

    # ========================================================================
    # Logs
    # ========================================================================

    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Get deployment logs from all pods.

        Args:
            deployment: The deployment.
            follow: Whether to follow logs (only follows first pod if true).
            tail: Number of lines to tail per pod.

        Yields:
            Log lines from all pods, prefixed with pod name.

        Raises:
            DeploymentLogsNotFoundError: If no Deployment found or pods unavailable.
            DeployerError: If deployment has no inventory.
        """
        ctx = self._initialize_deployment_context(deployment)

        stored_metadata = deployment.deployment_metadata or {}
        inventory_data = stored_metadata.get("resource_inventory")

        if not inventory_data:
            raise DeployerError(
                f"Deployment '{deployment.name}' has no resource inventory"
            )

        inventory = [ResourceInventoryItem(**item) for item in inventory_data]

        deployment_items = [
            item
            for item in inventory
            if item.kind == "Deployment" and item.api_version == "apps/v1"
        ]

        if not deployment_items:
            raise DeploymentLogsNotFoundError(
                f"No Deployment resources in inventory for '{deployment.name}'"
            )

        deployment_item = deployment_items[0]

        try:
            k8s_deployment_obj = self.k8s_applier.get_resource(
                name=deployment_item.name,
                namespace=deployment_item.namespace,
                kind="Deployment",
                api_version="apps/v1",
            )

            if not k8s_deployment_obj:
                raise DeploymentLogsNotFoundError(
                    f"Deployment '{deployment_item.name}' not found in cluster"
                )

            k8s_deployment = normalize_resource_to_dict(k8s_deployment_obj)
            label_selector: Dict[str, str] = {}
            try:
                spec = k8s_deployment.get("spec", {})
                selector = spec.get("selector", {})
                label_selector = (
                    selector.get("matchLabels")
                    or selector.get("match_labels")
                    or {}
                )
            except (AttributeError, TypeError):
                pass

            if not label_selector:
                raise DeploymentLogsNotFoundError(
                    f"Deployment '{deployment_item.name}' has no label selector"
                )

            label_selector_str = ",".join(
                f"{k}={v}" for k, v in label_selector.items()
            )

            pods = self.k8s_applier.list_resources(
                namespace=deployment_item.namespace or ctx.namespace,
                kind="Pod",
                api_version="v1",
                label_selector=label_selector_str,
            )

            if not pods:
                raise DeploymentLogsNotFoundError(
                    f"No pods found for Deployment '{deployment_item.name}'"
                )

            container_name = "main"
            try:
                spec = k8s_deployment.get("spec", {})
                template = spec.get("template", {})
                template_spec = template.get("spec", {})
                containers = template_spec.get("containers", [])
                if (
                    containers
                    and isinstance(containers, list)
                    and len(containers) > 0
                ):
                    container_name = containers[0].get("name", "main")
            except (AttributeError, TypeError, IndexError):
                pass

            if follow:
                pod_name = pods[0].metadata.name
                resp = self.k8s_core_api.read_namespaced_pod_log(
                    name=pod_name,
                    namespace=deployment_item.namespace or ctx.namespace,
                    container=container_name,
                    follow=True,
                    tail_lines=tail,
                    _preload_content=False,
                )
                for line in resp:
                    if isinstance(line, bytes):
                        yield line.decode("utf-8").rstrip()
                    else:
                        yield str(line).rstrip()
            else:
                for pod in pods:
                    pod_name = pod.metadata.name
                    try:
                        logs = self.k8s_core_api.read_namespaced_pod_log(
                            name=pod_name,
                            namespace=deployment_item.namespace
                            or ctx.namespace,
                            container=container_name,
                            tail_lines=tail,
                        )
                        for line in logs.split("\n"):
                            if line:
                                yield f"[{pod_name}] {line}"
                    except ApiException as pod_err:
                        logger.warning(
                            f"Failed to get logs from pod '{pod_name}': {pod_err}"
                        )
                        continue

        except ApiException as e:
            raise DeploymentLogsNotFoundError(
                f"Failed to get logs for '{deployment.name}': {e}"
            )

    # ========================================================================
    # Deprovisioning
    # ========================================================================

    def do_deprovision_deployment(
        self,
        deployment: DeploymentResponse,
        timeout: int,
    ) -> Optional[DeploymentOperationalState]:
        """Deprovision a deployment using inventory-based deletion.

        Args:
            deployment: The deployment to deprovision.
            timeout: Timeout in seconds.

        Returns:
            None to indicate immediate deletion.

        Raises:
            DeploymentDeprovisionError: If deprovisioning fails.
            DeploymentNotFoundError: If deployment has no inventory.
        """
        stored_metadata = deployment.deployment_metadata
        inventory_data = stored_metadata.get("resource_inventory")

        if not inventory_data:
            logger.info(
                f"No inventory found for deployment '{deployment.name}'. "
                f"Treating as already deprovisioned."
            )
            raise DeploymentNotFoundError(
                f"No resource inventory stored for deployment '{deployment.name}'"
            )

        inventory = [ResourceInventoryItem(**item) for item in inventory_data]
        logger.info(
            f"Deprovisioning deployment '{deployment.name}' "
            f"({len(inventory)} resource(s) in inventory)"
        )

        try:
            result = self.k8s_applier.delete_from_inventory(
                inventory=inventory,
                propagation_policy="Foreground",
            )

            if result.deleted_count == 0 and result.failed_count == 0:
                logger.info(
                    f"All resources for deployment '{deployment.name}' "
                    f"were already deleted or skipped"
                )
                return None

            if result.failed_count == 0:
                logger.info(
                    f"Successfully deprovisioned deployment '{deployment.name}' "
                    f"({result.deleted_count} deleted, {result.skipped_count} skipped)"
                )
                return None

            failed_list = "\n".join(
                f"  ❌ {r}" for r in result.failed_resources
            )

            error_msg = (
                f"Partial deprovisioning failure for deployment '{deployment.name}':\n"
                f"  ✅ Deleted: {result.deleted_count} resource(s)\n"
                f"  ⏭️  Skipped: {result.skipped_count} resource(s)\n"
                f"  ❌ Failed: {result.failed_count} resource(s)\n\n"
                f"Resources that failed to delete:\n{failed_list}\n\n"
                f"The deployment will NOT be removed from ZenML to allow retry or manual cleanup.\n"
                f"You can retry deletion with: zenml deployment delete {deployment.name}"
            )
            logger.error(error_msg)
            raise DeploymentDeprovisionError(error_msg)

        except DeploymentDeprovisionError:
            raise
        except Exception as e:
            logger.error(
                f"Unexpected error while deprovisioning '{deployment.name}': {e}. "
                f"Some resources may still exist. Manual cleanup may be required."
            )
            raise DeploymentDeprovisionError(
                f"Failed to deprovision deployment '{deployment.name}': {e}"
            ) from e



"""Add cached heartbeat threshold column [2837a7ab3d77].

Revision ID: 2837a7ab3d77
Revises: d203788f82b9
Create Date: 2025-11-28 12:27:14.341553

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2837a7ab3d77"
down_revision = "6e4eb89f632d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("heartbeat_threshold", sa.Integer(), nullable=True)
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.drop_column("heartbeat_threshold")



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML OAuth2 device authorization grant client support."""

import platform
import time
import webbrowser
from typing import Optional, Union
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

import requests

from zenml import __version__
from zenml.cli import utils as cli_utils
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import (
    API,
    AUTH,
    DEFAULT_HTTP_TIMEOUT,
    DEVICE_AUTHORIZATION,
    LOGIN,
    VERSION_1,
)
from zenml.exceptions import AuthorizationException, OAuthError
from zenml.logger import get_logger
from zenml.login.credentials import APIToken
from zenml.login.pro.constants import ZENML_PRO_API_URL

logger = get_logger(__name__)


def web_login(
    url: Optional[str] = None,
    verify_ssl: Optional[Union[str, bool]] = None,
    pro_api_url: Optional[str] = None,
    preferred_workspace: Optional[str] = None,
) -> APIToken:
    """Implements the OAuth2 Device Authorization Grant flow.

    This function implements the client side of the OAuth2 Device Authorization
    Grant flow as defined in https://tools.ietf.org/html/rfc8628, with the
    following customizations:

    * the unique ZenML client ID (`user_id` in the global config) is used
    as the OAuth2 client ID value
    * additional information is added to the user agent header to be used by
    users to identify the ZenML client

    Upon completion of the flow, the access token is saved in the credentials store.

    Args:
        url: The URL of the OAuth2 server. If not provided, the ZenML Pro API
            server is used by default.
        verify_ssl: Whether to verify the SSL certificate of the OAuth2 server.
            If a string is passed, it is interpreted as the path to a CA bundle
            file.
        pro_api_url: The URL of the ZenML Pro API server. If not provided, the
            default ZenML Pro API server URL is used.
        preferred_workspace: Optional ZenML Pro workspace hint (UUID or name)
            to pass to the browser verification URL.

    Returns:
        The response returned by the OAuth2 server.

    Raises:
        AuthorizationException: If an error occurred during the authorization
            process.
    """
    from zenml.login.credentials_store import get_credentials_store
    from zenml.models import (
        OAuthDeviceAuthorizationRequest,
        OAuthDeviceAuthorizationResponse,
        OAuthDeviceTokenRequest,
        OAuthDeviceUserAgentHeader,
        OAuthTokenResponse,
    )

    credentials_store = get_credentials_store()

    # Make a request to the OAuth2 server to get the device code and user code.
    # The client ID used for the request is the unique ID of the ZenML client.
    response: Optional[requests.Response] = None

    # Add the following information in the user agent header to be used by users
    # to identify the ZenML client:
    #
    # * the ZenML version
    # * the python version
    # * the OS type
    # * the hostname
    #
    user_agent_header = OAuthDeviceUserAgentHeader(
        hostname=platform.node(),
        zenml_version=__version__,
        python_version=platform.python_version(),
        os=platform.system(),
    )

    zenml_pro = False
    if not url:
        # If no URL is provided, we use the ZenML Pro API server by default
        zenml_pro = True
        url = base_url = pro_api_url or ZENML_PRO_API_URL
    else:
        # Get rid of any trailing slashes to prevent issues when having double
        # slashes in the URL
        url = url.rstrip("/")
        if pro_api_url:
            # This is a ZenML Pro server. The device authentication is done
            # through the ZenML Pro API.
            zenml_pro = True
            base_url = pro_api_url
        else:
            base_url = url

    auth_request = OAuthDeviceAuthorizationRequest(
        client_id=GlobalConfiguration().user_id
    )

    # If an existing token is found in the credentials store, we reuse its
    # device ID to avoid creating a new device ID for the same device.
    existing_token = credentials_store.get_token(url)
    if existing_token and existing_token.device_id:
        auth_request.device_id = existing_token.device_id

    if zenml_pro:
        auth_url = base_url + AUTH + DEVICE_AUTHORIZATION
        login_url = base_url + AUTH + LOGIN
    else:
        auth_url = base_url + API + VERSION_1 + DEVICE_AUTHORIZATION
        login_url = base_url + API + VERSION_1 + LOGIN

    try:
        response = requests.post(
            auth_url,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": user_agent_header.encode(),
            },
            data=auth_request.model_dump(exclude_none=True),
            verify=verify_ssl,
            timeout=DEFAULT_HTTP_TIMEOUT,
        )
        if response.status_code == 200:
            auth_response = OAuthDeviceAuthorizationResponse(**response.json())
        else:
            logger.info(f"Error: {response.status_code} {response.text}")
            raise AuthorizationException(
                f"Could not connect to {base_url}. Please check the URL."
            )
    except (requests.exceptions.JSONDecodeError, ValueError, TypeError):
        logger.exception("Bad response received from API server.")
        raise AuthorizationException(
            "Bad response received from API server. Please check the URL."
        )
    except requests.exceptions.RequestException:
        logger.exception("Could not connect to API server.")
        raise AuthorizationException(
            f"Could not connect to {base_url}. Please check the URL."
        )

    # Open the verification URL in the user's browser
    verification_uri = (
        auth_response.verification_uri_complete
        or auth_response.verification_uri
    )
    if verification_uri.startswith("/"):
        # If the verification URI is a relative path, we need to add the base
        # URL to it
        verification_uri = base_url + verification_uri

    # If a workspace identifier is provided, we add it to the verification URL
    # as a query parameter, so the dashboard can preselect it.
    if preferred_workspace:
        parsed_uri = urlparse(verification_uri)
        query_params = dict(parse_qsl(parsed_uri.query))
        query_params["workspace"] = preferred_workspace
        verification_uri = urlunparse(
            parsed_uri._replace(query=urlencode(query_params))
        )

    webbrowser.open(verification_uri)

    # Display the verification URL without panel styling
    from zenml.console import console

    console.print()
    console.print(
        "If your browser did not open automatically, please open the following URL into your browser to proceed with the authentication:",
        style="white",
    )
    console.print(verification_uri, style="bright_blue underline")
    console.print()

    # Poll the OAuth2 server until the user has authorized the device
    token_request = OAuthDeviceTokenRequest(
        device_code=auth_response.device_code,
        client_id=auth_request.client_id,
    )
    expires_in = auth_response.expires_in
    interval = auth_response.interval
    token_response: OAuthTokenResponse
    while True:
        response = requests.post(
            login_url,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data=token_request.model_dump(),
            verify=verify_ssl,
            timeout=DEFAULT_HTTP_TIMEOUT,
        )
        if response.status_code == 200:
            # The user has authorized the device, so we can extract the access token
            token_response = OAuthTokenResponse(**response.json())
            if zenml_pro:
                cli_utils.success("✔ Successfully logged in to ZenML Pro.")
            else:
                cli_utils.success(f"✔ Successfully logged in to {url}.")
            break
        elif response.status_code == 400:
            try:
                error_response = OAuthError(**response.json())
            except (
                requests.exceptions.JSONDecodeError,
                ValueError,
                TypeError,
            ):
                raise AuthorizationException(
                    f"Error received from {base_url}: {response.text}"
                )

            if error_response.error == "authorization_pending":
                # The user hasn't authorized the device yet, so we wait for the
                # interval and try again
                pass
            elif error_response.error == "slow_down":
                # The OAuth2 server is asking us to slow down our polling
                interval += 5
            else:
                # There was another error with the request
                raise AuthorizationException(
                    f"Error: {error_response.error} {error_response.error_description}"
                )

            expires_in -= interval
            if expires_in <= 0:
                raise AuthorizationException(
                    "User did not authorize the device in time."
                )
            time.sleep(interval)
        else:
            # There was another error with the request
            raise AuthorizationException(
                f"Error: {response.status_code} {response.text}"
            )

    # Save the token in the credentials store
    return credentials_store.set_token(
        url, token_response, is_zenml_pro=zenml_pro
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base class for log stores."""

import logging
import threading
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Type, cast

from zenml.enums import StackComponentType
from zenml.models import LogsResponse
from zenml.stack import Flavor, StackComponent, StackComponentConfig
from zenml.utils.logging_utils import LogEntry


class BaseLogStoreConfig(StackComponentConfig):
    """Base configuration for all log stores."""


class BaseLogStoreOrigin:
    """Base class for all ZenML log store origins.

    The origin is the entry point for all log records to be sent to the log
    store for processing. The process of sending a log record is as follows:

    1. instantiate the log store or use the active log store
    2. register an origin by calling log_store.register_origin() and passing
    the log model and optional metadata to be attached to each log record
    3. emit the log record by calling log_store.emit() and passing the origin
    and log record
    4. deregister the origin when all logs have been emitted by calling
    log_store.deregister(origin)
    """

    def __init__(
        self,
        name: str,
        log_store: "BaseLogStore",
        log_model: LogsResponse,
        metadata: Dict[str, Any],
    ) -> None:
        """Initialize a log store origin.

        Args:
            name: The name of the origin.
            log_store: The log store to emit logs to.
            log_model: The log model associated with the origin.
            metadata: Additional metadata to attach to all log entries that will
                be emitted by this origin.
        """
        self.name = name
        self.log_store = log_store
        self.log_model = log_model
        self.metadata = metadata


class BaseLogStore(StackComponent, ABC):
    """Base class for all ZenML log stores.

    A log store is responsible for collecting, storing, and retrieving logs
    during pipeline and step execution. Different implementations may store
    logs in different backends (artifact store, OpenTelemetry, Datadog, etc.).
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the log store.

        Args:
            *args: Positional arguments for the base class.
            **kwargs: Keyword arguments for the base class.
        """
        super().__init__(*args, **kwargs)
        self._origins: Dict[str, BaseLogStoreOrigin] = {}
        self._lock = threading.RLock()

    @property
    def config(self) -> BaseLogStoreConfig:
        """Returns the configuration of the log store.

        Returns:
            The configuration.
        """
        return cast(BaseLogStoreConfig, self._config)

    @property
    def origin_class(self) -> Type[BaseLogStoreOrigin]:
        """Class of the origin.

        Returns:
            The class of the origin used with this log store.
        """
        return BaseLogStoreOrigin

    def register_origin(
        self, name: str, log_model: LogsResponse, metadata: Dict[str, Any]
    ) -> BaseLogStoreOrigin:
        """Register an origin for the log store.

        Args:
            name: The name of the origin.
            log_model: The log model associated with the origin.
            metadata: Additional metadata to attach to the log entry.

        Returns:
            The origin.
        """
        with self._lock:
            origin = self.origin_class(name, self, log_model, metadata)
            self._origins[name] = origin
            return origin

    def deregister_origin(
        self,
        origin: BaseLogStoreOrigin,
        blocking: bool = True,
    ) -> None:
        """Deregister an origin previously registered with the log store.

        If no other origins are left, the log store will be flushed. The
        `blocking` parameter determines whether to block until the flush is
        complete.

        Args:
            origin: The origin to deregister.
            blocking: Whether to block until the deregistration is complete
                and all logs are flushed if this is the last origin registered.
        """
        with self._lock:
            if origin.name not in self._origins:
                return
            self._release_origin(origin)
            del self._origins[origin.name]
            if len(self._origins) == 0:
                self.flush(blocking=blocking)

    @abstractmethod
    def emit(
        self,
        origin: BaseLogStoreOrigin,
        record: logging.LogRecord,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Process a log record from the logging system.

        Args:
            origin: The origin used to send the log record.
            record: The Python logging.LogRecord to process.
            metadata: Additional metadata to attach to the log entry.
        """

    @abstractmethod
    def _release_origin(
        self,
        origin: BaseLogStoreOrigin,
    ) -> None:
        """Finalize the stream of log records associated with an origin.

        This is used to announce the end of the stream of log records associated
        with an origin and that no more log records will be emitted.

        The implementation should ensure that all log records associated with
        the origin are flushed to the backend and any resources (clients,
        connections, file descriptors, etc.) are released.

        Args:
            origin: The origin to finalize.
        """

    @abstractmethod
    def flush(self, blocking: bool = True) -> None:
        """Flush the log store.

        This method is called to ensure that all logs are flushed to the backend.

        Args:
            blocking: Whether to block until the flush is complete.
        """

    @abstractmethod
    def fetch(
        self,
        logs_model: LogsResponse,
        limit: int,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ) -> List[LogEntry]:
        """Fetch logs from the log store.

        This method is called from the server to retrieve logs for display
        on the dashboard or via API. The implementation should not require
        any integration-specific SDKs that aren't available on the server.

        Args:
            logs_model: The logs model containing metadata about the logs.
            start_time: Filter logs after this time.
            end_time: Filter logs before this time.
            limit: Maximum number of log entries to return.

        Returns:
            List of log entries matching the query.
        """


class BaseLogStoreFlavor(Flavor):
    """Base class for all ZenML log store flavors."""

    @property
    def type(self) -> StackComponentType:
        """Type of the flavor.

        Returns:
            The type of the flavor.
        """
        return StackComponentType.LOG_STORE

    @property
    def config_class(self) -> Type[BaseLogStoreConfig]:
        """Config class for the base log store flavor.

        Returns:
            The config class.
        """
        return BaseLogStoreConfig

    @property
    @abstractmethod
    def implementation_class(self) -> Type["BaseLogStore"]:
        """Implementation class for the base log store flavor.

        Returns:
            The implementation class.
        """



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Run:AI step operator flavor."""

from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Type
from urllib.parse import urlparse

from pydantic import (
    Field,
    PositiveFloat,
    PositiveInt,
    field_validator,
    model_validator,
)

from zenml.config.base_settings import BaseSettings
from zenml.integrations.runai import RUNAI_STEP_OPERATOR_FLAVOR
from zenml.step_operators import BaseStepOperatorConfig, BaseStepOperatorFlavor
from zenml.utils.secret_utils import PlainSerializedSecretStr

if TYPE_CHECKING:
    from zenml.integrations.runai.step_operators import RunAIStepOperator


class RunAIStepOperatorSettings(BaseSettings):
    """Per-step settings for Run:AI execution.

    These settings can be configured per-step using the step decorator:

    ```python
    @step(
        step_operator="runai",
        settings={"step_operator": {"gpu_portion_request": 0.5}}
    )
    def my_step():
        ...
    ```
    """

    gpu_devices_request: int = Field(
        default=1,
        ge=0,
        description="Number of GPUs to request for the workload. Default 1 "
        "requests a whole device; set to 0 for CPU-only or keep 1 with "
        "gpu_portion_request < 1.0 to request a fractional share. Examples: "
        "1 for single GPU, 2 for dual GPU, 0 for CPU-only. Pair with "
        "gpu_request_type='portion' to control fractional allocation.",
    )
    gpu_portion_request: float = Field(
        default=1.0,
        ge=0.0,
        le=1.0,
        description="Fractional GPU allocation as a decimal (0.0-1.0). "
        "Default 1.0 requests a full device; lower values (e.g., 0.5 or 0.25) "
        "share a GPU across workloads. Used when gpu_request_type='portion'.",
    )
    gpu_request_type: Literal["portion", "memory"] = Field(
        default="portion",
        description="GPU allocation method. 'portion' for fractional GPU sharing "
        "(use gpu_portion_request: 1.0 for full GPU, 0.5 for half). "
        "'memory' for exact memory allocation (gpu_memory_request)",
    )
    gpu_memory_request: Optional[str] = Field(
        default=None,
        description="GPU memory to request (e.g., '20Gi'). Used when "
        "gpu_request_type='memory'. Provides consistent memory allocation "
        "across heterogeneous GPU types",
    )

    gpu_portion_limit: Optional[float] = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Maximum GPU portion limit for dynamic fractions. Enables burst: "
        "workload gets gpu_portion_request guaranteed, can use up to this limit when available",
    )
    gpu_memory_limit: Optional[str] = Field(
        default=None,
        description="Maximum GPU memory limit for dynamic fractions (e.g., '40Gi'). "
        "Used with gpu_memory_request for burst capacity",
    )

    cpu_core_request: float = Field(
        default=1.0,
        ge=0.1,
        description="Number of CPU cores to request. Examples: 1.0 for one core, "
        "2.5 for two and a half cores, 0.5 for half a core",
    )
    cpu_core_limit: Optional[float] = Field(
        default=None,
        ge=0.1,
        description="Maximum CPU cores limit. If set, allows bursting from "
        "cpu_core_request to this limit",
    )
    cpu_memory_request: str = Field(
        default="4G",
        description="RAM to request for the workload container. Must use Kubernetes "
        "format with valid units: K, M, G, T, Ki, Mi, Gi, Ti. Examples: '4G' "
        "for 4 gigabytes, '512M' for 512 megabytes, '8Gi' for 8 gibibytes. "
        "Default is sufficient for most training jobs.",
    )
    cpu_memory_limit: Optional[str] = Field(
        default=None,
        description="Maximum memory limit (e.g., '16Gi'). If set, allows bursting "
        "from cpu_memory_request",
    )

    node_pools: Optional[List[str]] = Field(
        default=None,
        description="Ordered list of node pool names for scheduling. Scheduler tries pools "
        "in order. Node pools group nodes by GPU type, InfiniBand, etc. "
        "Example: ['a100-pool', 'v100-pool']",
    )
    node_type: Optional[str] = Field(
        default=None,
        description="Node type label for GPU selection (e.g., 'A100', 'V100'). "
        "Alternative to node_pools for simpler GPU type selection",
    )

    preemptibility: Optional[Literal["preemptible", "non-preemptible"]] = (
        Field(
            default=None,
            description="Workload preemption policy. 'preemptible' allows over-quota usage "
            "but may be preempted by higher priority workloads. 'non-preemptible' guarantees "
            "resources within quota",
        )
    )
    priority_class: Optional[str] = Field(
        default=None,
        description="Kubernetes PriorityClass name for scheduling priority within project queue. "
        "Run:AI presets: 'train' (50), 'build' (50), 'interactive-preemptible' (100), 'inference' (125)",
    )

    tolerations: Optional[List[Dict[str, Any]]] = Field(
        default=None,
        description="Kubernetes tolerations for scheduling on tainted nodes. Each item: "
        "{'key': str, 'operator': 'Equal'|'Exists', 'value': str, "
        "'effect': 'NoSchedule'|'PreferNoSchedule'|'NoExecute'}. "
        "Example: [{'key': 'nvidia.com/gpu', 'operator': 'Exists', 'effect': 'NoSchedule'}]",
    )

    extended_resources: Optional[Dict[str, str]] = Field(
        default=None,
        description="Extended Kubernetes resources (InfiniBand, FPGAs, etc). "
        "Format: {resource_name: quantity}. Example: {'nvidia.com/mlnx-nic': '1', 'rdma/hca': '1'}",
    )
    large_shm_request: bool = Field(
        default=False,
        description="Request large /dev/shm (shared memory). Required for PyTorch DataLoader "
        "with num_workers > 0, or multi-process training that uses shared memory",
    )

    backoff_limit: Optional[int] = Field(
        default=None,
        ge=0,
        description="Number of retries before marking workload as failed. Default is 6 in "
        "Kubernetes. Set to 0 for no retries",
    )
    termination_grace_period_seconds: Optional[int] = Field(
        default=None,
        ge=0,
        description="Seconds to wait for graceful termination before force killing. "
        "Default is 30 seconds. Increase for workloads needing cleanup time",
    )
    terminate_after_preemption: Optional[bool] = Field(
        default=None,
        description="If True, terminate workload after preemption instead of requeuing. "
        "Default is False (workload gets rescheduled)",
    )
    working_dir: Optional[str] = Field(
        default=None,
        description="Working directory inside container. If not set, uses container's default",
    )

    labels: Optional[Dict[str, str]] = Field(
        default=None,
        description="Kubernetes labels to add to the workload pod. "
        "Example: {'team': 'ml-research', 'cost-center': 'ai-division'}",
    )
    annotations: Optional[Dict[str, str]] = Field(
        default=None,
        description="Kubernetes annotations to add to the workload pod. "
        "Example: {'prometheus.io/scrape': 'true'}",
    )

    workload_timeout: Optional[PositiveInt] = Field(
        default=None,
        description="Maximum time in seconds to wait for workload completion. "
        "If the workload doesn't finish within this time, it will be marked as failed. "
        "If not set, workloads can run indefinitely. Example: 3600 for 1 hour",
    )
    pending_timeout: Optional[PositiveInt] = Field(
        default=None,
        description="Maximum time in seconds a workload may remain in the pending state "
        "before it is stopped. If not set, workloads can remain pending indefinitely. "
        "Example: 600 for 10 minutes",
    )

    @field_validator(
        "cpu_memory_request",
        "cpu_memory_limit",
        "gpu_memory_request",
        "gpu_memory_limit",
    )
    @classmethod
    def _validate_memory_format(cls, value: Optional[str]) -> Optional[str]:
        """Validates memory request/limit format.

        Args:
            value: The memory value.

        Returns:
            The validated value.

        Raises:
            ValueError: If the memory format is invalid.
        """
        if value is None:
            return None

        import re

        # Use regex for stricter validation
        pattern = r"^(\d+(?:\.\d+)?)(K|M|G|T|Ki|Mi|Gi|Ti)$"
        if not re.match(pattern, value):
            raise ValueError(
                f"Invalid memory format: {value}. "
                f"Must be a number followed by one of: K, M, G, T, Ki, Mi, Gi, Ti. "
                f"Examples: '4G', '512M', '8Gi'"
            )

        return value

    @model_validator(mode="after")
    def _validate_gpu_settings(self) -> "RunAIStepOperatorSettings":
        """Validate GPU settings consistency.

        Returns:
            The validated settings.

        Raises:
            ValueError: If GPU settings are inconsistent.
        """
        if self.gpu_request_type == "memory" and not self.gpu_memory_request:
            raise ValueError(
                "gpu_memory_request is required when gpu_request_type='memory'"
            )
        if self.gpu_portion_limit is not None:
            if self.gpu_portion_limit < self.gpu_portion_request:
                raise ValueError(
                    "gpu_portion_limit must be >= gpu_portion_request"
                )
        return self


class RunAIStepOperatorConfig(
    BaseStepOperatorConfig, RunAIStepOperatorSettings
):
    """Configuration for the Run:AI step operator.

    This step operator enables running individual pipeline steps on Run:AI
    clusters with fractional GPU allocation.

    Example stack configuration:
    ```bash
    zenml step-operator register runai \\
        --flavor=runai \\
        --client_id="xxx" \\
        --client_secret="xxx" \\
        --runai_base_url="https://myorg.run.ai" \\
        --project_name="my-project"
    ```
    """

    client_id: PlainSerializedSecretStr = Field(
        ...,
        description="Run:AI client ID for API authentication. Obtain from Run:AI "
        "control plane under Settings > Application > Applications. Required for API access",
    )
    client_secret: PlainSerializedSecretStr = Field(
        ...,
        description="Run:AI client secret for API authentication. Obtain from Run:AI "
        "control plane under Settings > Application > Applications.",
    )
    runai_base_url: str = Field(
        ...,
        description="Run:AI control plane base URL. For Run:AI SaaS, use format "
        "'https://<organization>.run.ai'. For self-hosted Run:AI deployments, use your control plane URL. "
        "Example: 'https://my-org.run.ai'",
    )

    @field_validator("runai_base_url")
    @classmethod
    def _validate_runai_base_url(cls, value: str) -> str:
        """Validate and normalize Run:AI base URL.

        Args:
            value: The URL to validate.

        Returns:
            The validated and normalized URL.

        Raises:
            ValueError: If the URL is invalid.
        """
        if not value.startswith(("http://", "https://")):
            raise ValueError(
                f"Invalid URL '{value}'. Must start with http:// or https://"
            )

        parsed = urlparse(value)
        if not parsed.hostname:
            raise ValueError(
                f"Invalid URL '{value}'. URL must include a hostname."
            )

        if parsed.query or parsed.fragment:
            raise ValueError(
                f"Invalid URL '{value}'. Query parameters and fragments are "
                "not supported in runai_base_url."
            )

        return value.rstrip("/")

    project_name: str = Field(
        ...,
        description="Run:AI project name for workload submission. The project must "
        "exist in the Run:AI control plane and have sufficient quota. Workloads are "
        "billed against this project's resources",
    )
    cluster_name: Optional[str] = Field(
        default=None,
        description="Run:AI cluster name for workload execution. If not specified, "
        "uses the cluster associated with the project or the first available cluster",
    )

    image_pull_secret_name: Optional[str] = Field(
        default=None,
        description="Name of an existing Run:AI image pull secret for private registries. "
        "Create the secret in Run:AI UI under Credentials > Docker Registry. "
        "If not specified, assumes the container registry is public or credentials "
        "are pre-configured in the Run:AI project",
    )

    monitoring_interval: PositiveFloat = Field(
        default=30.0,
        description="Interval in seconds to poll Run:AI API for workload status. "
        "Lower values provide faster status updates but increase API load. "
        "Recommended: 30-60 seconds for production",
    )
    delete_on_failure: bool = Field(
        default=False,
        description="Whether to delete Run:AI workloads after they fail or timeout. "
        "Workloads are always stopped on failure to halt resource usage; "
        "set this to True to delete them afterward. Defaults to False to preserve "
        "failed workloads for traceability in the Run:AI UI.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        Run:AI step operator always runs remotely on Run:AI clusters.

        Returns:
            bool: Always ``True`` because the Run:AI step operator runs remotely.
        """
        return True

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Run:AI step operator never runs locally.

        Returns:
            bool: Always ``False`` because the Run:AI step operator is remote.
        """
        return False


class RunAIStepOperatorFlavor(BaseStepOperatorFlavor):
    """Run:AI step operator flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return RUNAI_STEP_OPERATOR_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the flavor.

        Returns:
            The display name of the flavor.
        """
        return "Run:AI"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/step_operator/runai.png"

    @property
    def config_class(self) -> Type[RunAIStepOperatorConfig]:
        """Returns RunAIStepOperatorConfig config class.

        Returns:
            The config class.
        """
        return RunAIStepOperatorConfig

    @property
    def implementation_class(self) -> Type["RunAIStepOperator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.runai.step_operators import RunAIStepOperator

        return RunAIStepOperator



"""Dynamic pipelines [af27025fe19c].

Revision ID: af27025fe19c
Revises: 0.91.0
Create Date: 2025-10-27 13:23:40.442485

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "af27025fe19c"
down_revision = "0.91.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("is_dynamic", sa.Boolean(), nullable=True)
        )

    op.execute(
        "UPDATE pipeline_snapshot SET is_dynamic = FALSE WHERE is_dynamic IS NULL"
    )
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.alter_column(
            "is_dynamic", existing_type=sa.Boolean(), nullable=False
        )

    with op.batch_alter_table("step_configuration", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "step_run_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.alter_column(
            "snapshot_id", existing_type=sa.CHAR(length=32), nullable=True
        )
        batch_op.create_unique_constraint(
            "unique_step_configuration_for_snapshot_or_step_run",
            ["snapshot_id", "step_run_id", "name"],
        )
        batch_op.drop_constraint(
            "unique_step_name_for_snapshot", type_="unique"
        )
        batch_op.create_foreign_key(
            "fk_step_configuration_step_run_id_step_run",
            "step_run",
            ["step_run_id"],
            ["id"],
            ondelete="CASCADE",
        )

        batch_op.create_check_constraint(
            "ck_step_configuration_snapshot_step_run_exclusivity",
            "((snapshot_id IS NULL AND step_run_id IS NOT NULL) OR "
            "(snapshot_id IS NOT NULL AND step_run_id IS NULL))",
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "ck_step_configuration_snapshot_step_run_exclusivity",
        "step_configuration",
        type_="check",
    )
    with op.batch_alter_table("step_configuration", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_step_configuration_step_run_id_step_run", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "unique_step_name_for_snapshot", type_="unique"
        )
        batch_op.create_unique_constraint(
            "unique_step_name_for_snapshot", ["snapshot_id", "name"]
        )
        batch_op.alter_column(
            "snapshot_id", existing_type=sa.CHAR(length=32), nullable=False
        )
        batch_op.drop_column("step_run_id")

    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.drop_column("is_dynamic")

    # ### end Alembic commands ###



"""Release [0.92.0].

Revision ID: 0.92.0
Revises: d203788f82b9
Create Date: 2025-12-02 01:27:33.886780

"""

# revision identifiers, used by Alembic.
revision = "0.92.0"
down_revision = "d203788f82b9"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Kubernetes related functions.

Internal interface: no backwards compatibility guarantees.
Adjusted from https://github.com/tensorflow/tfx/blob/master/tfx/utils/kube_utils.py.
"""

import enum
import functools
import json
import re
import time
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from kubernetes import client as k8s_client
from kubernetes import config as k8s_config
from kubernetes.client.rest import ApiException
from urllib3.exceptions import ReadTimeoutError

from zenml.config.resource_settings import ByteUnit
from zenml.integrations.kubernetes.constants import (
    STEP_NAME_ANNOTATION_KEY,
)
from zenml.integrations.kubernetes.manifest_utils import (
    build_namespace_manifest,
    build_role_binding_manifest_for_service_account,
    build_secret_manifest,
    build_service_account_manifest,
)
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.integrations.kubernetes.serialization_utils import (
    normalize_resource_to_dict,
)
from zenml.logger import get_logger
from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from zenml.config.resource_settings import ResourceSettings

logger = get_logger(__name__)

R = TypeVar("R")


# This is to fix a bug in the kubernetes client which has some wrong
# client-side validations that means the `on_exit_codes` field is
# unusable. See https://github.com/kubernetes-client/python/issues/2056
class PatchedFailurePolicyRule(k8s_client.V1PodFailurePolicyRule):  # type: ignore[misc]
    """Patched failure policy rule."""

    @property
    def on_pod_conditions(self):  # type: ignore[no-untyped-def]
        """On pod conditions.

        Returns:
            On pod conditions.
        """
        return self._on_pod_conditions

    @on_pod_conditions.setter
    def on_pod_conditions(self, on_pod_conditions):  # type: ignore[no-untyped-def]
        """On pod conditions.

        Args:
            on_pod_conditions: On pod conditions.
        """
        self._on_pod_conditions = on_pod_conditions


k8s_client.V1PodFailurePolicyRule = PatchedFailurePolicyRule
k8s_client.models.V1PodFailurePolicyRule = PatchedFailurePolicyRule


class PodPhase(enum.Enum):
    """Phase of the Kubernetes pod.

    Pod phases are defined in
    https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#pod-phase.
    """

    PENDING = "Pending"
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    UNKNOWN = "Unknown"


class JobStatus(enum.Enum):
    """Status of a Kubernetes job."""

    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"


def is_inside_kubernetes() -> bool:
    """Check whether we are inside a Kubernetes cluster or on a remote host.

    Returns:
        True if inside a Kubernetes cluster, else False.
    """
    try:
        k8s_config.load_incluster_config()
        return True
    except k8s_config.ConfigException:
        return False


def load_kube_config(
    incluster: bool = False, context: Optional[str] = None
) -> None:
    """Load the Kubernetes client config.

    Args:
        incluster: Whether to load the in-cluster config.
        context: Name of the Kubernetes context. If not provided, uses the
            currently active context. Will be ignored if `incluster` is True.
    """
    if incluster:
        k8s_config.load_incluster_config()
    else:
        k8s_config.load_kube_config(context=context)


def sanitize_label(label: str) -> str:
    """Sanitize a label for a Kubernetes resource.

    Args:
        label: The label to sanitize.

    Returns:
        The sanitized label.
    """
    # https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#rfc-1035-label-names
    label = re.sub(r"[^a-z0-9-]", "-", label.lower())
    label = re.sub(r"^[-]+", "", label)
    label = re.sub(r"[-]+", "-", label)
    label = label[:63]
    # Remove trailing dashes after truncation to make sure we end with an
    # alphanumeric character
    label = re.sub(r"[-]+$", "", label)

    return label


def pod_is_not_pending(pod: k8s_client.V1Pod) -> bool:
    """Check if pod status is not 'Pending'.

    Args:
        pod: Kubernetes pod.

    Returns:
        False if the pod status is 'Pending' else True.
    """
    return pod.status.phase != PodPhase.PENDING.value  # type: ignore[no-any-return]


def pod_failed(pod: k8s_client.V1Pod) -> bool:
    """Check if pod status is 'Failed'.

    Args:
        pod: Kubernetes pod.

    Returns:
        True if pod status is 'Failed' else False.
    """
    return pod.status.phase == PodPhase.FAILED.value  # type: ignore[no-any-return]


def pod_is_done(pod: k8s_client.V1Pod) -> bool:
    """Check if pod status is 'Succeeded'.

    Args:
        pod: Kubernetes pod.

    Returns:
        True if pod status is 'Succeeded' else False.
    """
    return pod.status.phase == PodPhase.SUCCEEDED.value  # type: ignore[no-any-return]


def get_pod(
    core_api: k8s_client.CoreV1Api,
    pod_name: str,
    namespace: str,
    api_request_timeout: Optional[int] = None,
) -> Optional[k8s_client.V1Pod]:
    """Get a pod from Kubernetes metadata API.

    Args:
        core_api: Client of `CoreV1Api` of Kubernetes API.
        pod_name: The name of the pod.
        namespace: The namespace of the pod.
        api_request_timeout: The request timeout in seconds.

    Raises:
        RuntimeError: When it sees unexpected errors from Kubernetes API.

    Returns:
        The found pod object. None if it's not found.
    """
    try:
        return retry_on_api_exception(
            core_api.read_namespaced_pod,
            api_request_timeout=api_request_timeout,
        )(name=pod_name, namespace=namespace)
    except k8s_client.rest.ApiException as e:
        if e.status == 404:
            return None
        raise RuntimeError from e


def wait_pod(
    kube_client_fn: Callable[[], k8s_client.ApiClient],
    pod_name: str,
    namespace: str,
    exit_condition_lambda: Callable[[k8s_client.V1Pod], bool],
    timeout_sec: int = 0,
    exponential_backoff: bool = False,
    stream_logs: bool = False,
    api_request_timeout: Optional[int] = None,
) -> k8s_client.V1Pod:
    """Wait for a pod to meet an exit condition.

    Args:
        kube_client_fn: the kube client fn is a function that is called
            periodically and is used to get a `CoreV1Api` client for
            the Kubernetes API. It should cache the client to avoid
            unnecessary overhead but should also instantiate a new client if
            the previous one is using credentials that are about to expire.
        pod_name: The name of the pod.
        namespace: The namespace of the pod.
        exit_condition_lambda: A lambda
            which will be called periodically to wait for a pod to exit. The
            function returns True to exit.
        timeout_sec: Timeout in seconds to wait for pod to reach exit
            condition, or 0 to wait for an unlimited duration.
            Defaults to unlimited.
        exponential_backoff: Whether to use exponential back off for polling.
            Defaults to False.
        stream_logs: Whether to stream the pod logs to
            `zenml.logger.info()`. Defaults to False.
        api_request_timeout: The request timeout in seconds.

    Raises:
        RuntimeError: when the function times out.

    Returns:
        The pod object which meets the exit condition.
    """
    start_time = utc_now()

    # Link to exponential back-off algorithm used here:
    # https://cloud.google.com/storage/docs/exponential-backoff
    backoff_interval = 1
    maximum_backoff = 32

    logged_lines = 0

    while True:
        kube_client = kube_client_fn()
        core_api = k8s_client.CoreV1Api(kube_client)

        resp = get_pod(
            core_api,
            pod_name,
            namespace,
            api_request_timeout=api_request_timeout,
        )

        if resp is None:
            raise RuntimeError(f"Pod `{namespace}:{pod_name}` not found.")

        # Stream logs to `zenml.logger.info()`.
        # TODO: can we do this without parsing all logs every time?
        if stream_logs and pod_is_not_pending(resp):
            try:
                response = core_api.read_namespaced_pod_log(
                    name=pod_name,
                    namespace=namespace,
                    _preload_content=False,
                )
            except ApiException as e:
                logger.error(f"Error reading pod logs: {e}. Retrying...")
            else:
                raw_data = response.data
                decoded_log = raw_data.decode("utf-8", errors="replace")
                logs = decoded_log.splitlines()
                if len(logs) > logged_lines:
                    for line in logs[logged_lines:]:
                        logger.info(line)
                    logged_lines = len(logs)

        # Raise an error if the pod failed.
        if pod_failed(resp):
            raise RuntimeError(f"Pod `{namespace}:{pod_name}` failed.")

        # Check if pod is in desired state (e.g. finished / running / ...).
        if exit_condition_lambda(resp):
            return resp

        # Check if wait timed out.
        elapse_time = utc_now() - start_time
        if elapse_time.seconds >= timeout_sec and timeout_sec != 0:
            raise RuntimeError(
                f"Waiting for pod `{namespace}:{pod_name}` timed out after "
                f"{timeout_sec} seconds."
            )

        # Wait (using exponential backoff).
        time.sleep(backoff_interval)
        if exponential_backoff and backoff_interval < maximum_backoff:
            backoff_interval *= 2


FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def _if_not_exists(create_fn: FuncT) -> FuncT:
    """Wrap a Kubernetes function to handle creation if already exists.

    Args:
        create_fn: Kubernetes function to be wrapped.

    Returns:
        Wrapped Kubernetes function.
    """

    def create_if_not_exists(*args: Any, **kwargs: Any) -> None:
        try:
            create_fn(*args, **kwargs)
        except ApiException as exc:
            if exc.status != 409:
                raise
            logger.debug(
                f"Didn't execute {create_fn.__name__} because already exists."
            )

    return cast(FuncT, create_if_not_exists)


def create_edit_service_account(
    core_api: k8s_client.CoreV1Api,
    rbac_api: k8s_client.RbacAuthorizationV1Api,
    service_account_name: str,
    namespace: str,
    role_binding_name: str = "zenml-edit",
) -> None:
    """Create a new Kubernetes service account with "edit" rights.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        rbac_api: Client of Rbac Authorization V1 API of Kubernetes API.
        service_account_name: Name of the service account.
        namespace: Kubernetes namespace. Defaults to "default".
        role_binding_name: Name of the role binding. Defaults to "zenml-edit".
    """
    rb_manifest = build_role_binding_manifest_for_service_account(
        name=role_binding_name,
        role_name="edit",
        service_account_name=service_account_name,
        namespace=namespace,
    )
    _if_not_exists(rbac_api.create_namespaced_role_binding)(
        namespace=namespace, body=rb_manifest
    )

    sa_manifest = build_service_account_manifest(
        name=service_account_name, namespace=namespace
    )
    _if_not_exists(core_api.create_namespaced_service_account)(
        namespace=namespace,
        body=sa_manifest,
    )


def create_namespace(core_api: k8s_client.CoreV1Api, namespace: str) -> None:
    """Create a Kubernetes namespace.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        namespace: Kubernetes namespace. Defaults to "default".
    """
    manifest = build_namespace_manifest(namespace)
    _if_not_exists(core_api.create_namespace)(body=manifest)


def create_secret(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    secret_name: str,
    data: Dict[str, Optional[str]],
) -> None:
    """Create a Kubernetes secret.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        namespace: The namespace in which to create the secret.
        secret_name: The name of the secret to create.
        data: The secret data.
    """
    core_api.create_namespaced_secret(
        namespace=namespace,
        body=build_secret_manifest(name=secret_name, data=data),
    )


def update_secret(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    secret_name: str,
    data: Dict[str, Optional[str]],
) -> None:
    """Update a Kubernetes secret.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        namespace: The namespace in which to update the secret.
        secret_name: The name of the secret to update.
        data: The secret data. If the value is None, the key will be removed
            from the secret.
    """
    core_api.patch_namespaced_secret(
        namespace=namespace,
        name=secret_name,
        body=build_secret_manifest(name=secret_name, data=data),
    )


def create_or_update_secret(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    secret_name: str,
    data: Dict[str, Optional[str]],
) -> None:
    """Create a Kubernetes secret if it doesn't exist, or update it if it does.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        namespace: The namespace in which to create or update the secret.
        secret_name: The name of the secret to create or update.
        data: The secret data. If the value is None, the key will be removed
            from the secret.

    Raises:
        ApiException: If the secret creation failed for any reason other than
            the secret already existing.
    """
    try:
        create_secret(core_api, namespace, secret_name, data)
    except ApiException as e:
        if e.status != 409:
            raise
        update_secret(core_api, namespace, secret_name, data)


def delete_secret(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    secret_name: str,
) -> None:
    """Delete a Kubernetes secret.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        namespace: The namespace in which to delete the secret.
        secret_name: The name of the secret to delete.
    """
    core_api.delete_namespaced_secret(
        name=secret_name,
        namespace=namespace,
    )


def create_and_wait_for_pod_to_start(
    core_api: k8s_client.CoreV1Api,
    pod_display_name: str,
    pod_name: str,
    pod_manifest: k8s_client.V1Pod,
    namespace: str,
    startup_max_retries: int,
    startup_failure_delay: float,
    startup_failure_backoff: float,
    startup_timeout: float,
    api_request_timeout: Optional[int] = None,
) -> None:
    """Create a pod and wait for it to reach a desired state.

    Args:
        core_api: Client of Core V1 API of Kubernetes API.
        pod_display_name: The display name of the pod to use in logs.
        pod_name: The name of the pod to create.
        pod_manifest: The manifest of the pod to create.
        namespace: The namespace in which to create the pod.
        startup_max_retries: The maximum number of retries for the pod startup.
        startup_failure_delay: The delay between retries for the pod startup.
        startup_failure_backoff: The backoff factor for the pod startup.
        startup_timeout: The maximum time to wait for the pod to start.
        api_request_timeout: The request timeout in seconds.

    Raises:
        TimeoutError: If the pod is still in a pending state after the maximum
            wait time has elapsed.
        Exception: If the pod fails to start after the maximum number of
            retries.
    """
    retries = 0

    while retries < startup_max_retries:
        try:
            # Create and run pod.
            core_api.create_namespaced_pod(
                namespace=namespace,
                body=pod_manifest,
            )
            break
        except Exception as e:
            retries += 1
            if retries < startup_max_retries:
                logger.debug(f"The {pod_display_name} failed to start: {e}")
                message = ""
                try:
                    if isinstance(e, ApiException) and e.body:
                        exception_body = json.loads(e.body)
                        message = exception_body.get("message", "")
                except Exception:
                    pass
                logger.error(
                    f"Failed to create {pod_display_name}. "
                    f"Retrying in {startup_failure_delay} seconds..."
                    "\nReason: " + message
                    if message
                    else ""
                )
                time.sleep(startup_failure_delay)
                startup_failure_delay *= startup_failure_backoff
            else:
                logger.error(
                    f"Failed to create {pod_display_name} after "
                    f"{startup_max_retries} retries. Exiting."
                )
                raise

    # Wait for pod to start
    logger.info(f"Waiting for {pod_display_name} to start...")
    max_wait = startup_timeout
    total_wait: float = 0
    delay = startup_failure_delay
    while True:
        pod = get_pod(
            core_api=core_api,
            pod_name=pod_name,
            namespace=namespace,
            api_request_timeout=api_request_timeout,
        )
        if not pod or pod_is_not_pending(pod):
            break
        if total_wait >= max_wait:
            # Have to delete the pending pod so it doesn't start running
            # later on.
            try:
                core_api.delete_namespaced_pod(
                    name=pod_name,
                    namespace=namespace,
                )
            except Exception:
                pass
            raise TimeoutError(
                f"The {pod_display_name} is still in a pending state "
                f"after {total_wait} seconds. Exiting."
            )

        if total_wait + delay > max_wait:
            delay = max_wait - total_wait
        total_wait += delay
        time.sleep(delay)
        delay *= startup_failure_backoff


def get_pod_owner_references(
    core_api: k8s_client.CoreV1Api,
    pod_name: str,
    namespace: str,
    api_request_timeout: Optional[int] = None,
) -> List[k8s_client.V1OwnerReference]:
    """Get owner references for a pod.

    Args:
        core_api: Kubernetes CoreV1Api client.
        pod_name: Name of the pod.
        namespace: Kubernetes namespace.
        api_request_timeout: The request timeout in seconds.

    Returns:
        List of owner references.
    """
    pod = get_pod(
        core_api=core_api,
        pod_name=pod_name,
        namespace=namespace,
        api_request_timeout=api_request_timeout,
    )

    if not pod or not pod.metadata or not pod.metadata.owner_references:
        return []

    return cast(
        List[k8s_client.V1OwnerReference], pod.metadata.owner_references
    )


def retry_on_api_exception(
    func: Callable[..., R],
    max_retries: int = 3,
    delay: float = 1,
    backoff: float = 1,
    fail_on_status_codes: Tuple[int, ...] = (404,),
    api_request_timeout: Optional[int] = None,
) -> Callable[..., R]:
    """Retry a function on API exceptions.

    Args:
        func: The function to retry.
        max_retries: The maximum number of retries.
        delay: The delay between retries.
        backoff: The backoff factor.
        fail_on_status_codes: The status codes to fail on immediately.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The wrapped function with retry logic.
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> R:
        _delay = delay
        retries = 0
        while retries <= max_retries:
            try:
                if api_request_timeout and "_request_timeout" not in kwargs:
                    kwargs["_request_timeout"] = api_request_timeout
                return func(*args, **kwargs)
            except (ApiException, ReadTimeoutError) as e:
                if (
                    isinstance(e, ApiException)
                    and e.status in fail_on_status_codes
                ):
                    raise

                retries += 1
                if retries <= max_retries:
                    func_name = getattr(func, "__name__", repr(func))
                    logger.warning("Error calling %s: %s.", func_name, e)
                    time.sleep(_delay)
                    _delay *= backoff
                else:
                    raise

        func_name = getattr(func, "__name__", repr(func))
        raise RuntimeError(
            f"Failed to call {func_name} after {max_retries} retries."
        )

    return wrapper


def create_job(
    batch_api: k8s_client.BatchV1Api,
    namespace: str,
    job_manifest: k8s_client.V1Job,
    api_request_timeout: Optional[int] = None,
) -> None:
    """Create a Kubernetes job.

    Args:
        batch_api: Kubernetes batch api.
        namespace: Kubernetes namespace.
        job_manifest: The manifest of the job to create.
        api_request_timeout: The request timeout in seconds.
    """
    retry_on_api_exception(
        batch_api.create_namespaced_job,
        api_request_timeout=api_request_timeout,
    )(
        namespace=namespace,
        body=job_manifest,
    )


def get_job(
    batch_api: k8s_client.BatchV1Api,
    namespace: str,
    job_name: str,
    api_request_timeout: Optional[int] = None,
) -> k8s_client.V1Job:
    """Get a job by name.

    Args:
        batch_api: Kubernetes batch api.
        namespace: Kubernetes namespace.
        job_name: The name of the job to get.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The job.
    """
    return retry_on_api_exception(
        batch_api.read_namespaced_job, api_request_timeout=api_request_timeout
    )(name=job_name, namespace=namespace)


def list_jobs(
    batch_api: k8s_client.BatchV1Api,
    namespace: str,
    label_selector: Optional[str] = None,
    api_request_timeout: Optional[int] = None,
) -> k8s_client.V1JobList:
    """List jobs in a namespace.

    Args:
        batch_api: Kubernetes batch api.
        namespace: Kubernetes namespace.
        label_selector: The label selector to use.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The job list.
    """
    return retry_on_api_exception(
        batch_api.list_namespaced_job, api_request_timeout=api_request_timeout
    )(
        namespace=namespace,
        label_selector=label_selector,
    )


def update_job(
    batch_api: k8s_client.BatchV1Api,
    namespace: str,
    job_name: str,
    annotations: Dict[str, str],
    api_request_timeout: Optional[int] = None,
) -> k8s_client.V1Job:
    """Update a job.

    Args:
        batch_api: Kubernetes batch api.
        namespace: Kubernetes namespace.
        job_name: The name of the job to update.
        annotations: The annotations to update.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The updated job.
    """
    return retry_on_api_exception(
        batch_api.patch_namespaced_job, api_request_timeout=api_request_timeout
    )(
        name=job_name,
        namespace=namespace,
        body={"metadata": {"annotations": annotations}},
    )


def is_step_job(job: k8s_client.V1Job) -> bool:
    """Check if a job is a step job.

    Args:
        job: The job to check.

    Returns:
        Whether the job is a step job.
    """
    if not job.metadata or not job.metadata.annotations:
        return False

    return STEP_NAME_ANNOTATION_KEY in job.metadata.annotations


def get_container_status(
    pod: k8s_client.V1Pod, container_name: str
) -> Optional[k8s_client.V1ContainerState]:
    """Get the status of a container.

    Args:
        pod: The pod to get the container status for.
        container_name: The container name.

    Returns:
        The container status.
    """
    if not pod.status or not pod.status.container_statuses:
        return None

    for container_status in pod.status.container_statuses:
        if container_status.name == container_name:
            return container_status.state

    return None


def get_container_termination_reason(
    pod: k8s_client.V1Pod, container_name: str
) -> Optional[Tuple[int, str]]:
    """Get the termination reason for a container.

    Args:
        pod: The pod to get the termination reason for.
        container_name: The container name.

    Returns:
        The exit code and termination reason for the container.
    """
    container_state = get_container_status(pod, container_name)
    if not container_state or not container_state.terminated:
        return None

    return (
        container_state.terminated.exit_code,
        container_state.terminated.reason or "Unknown",
    )


def wait_for_job_to_finish(
    get_client: Callable[[], k8s_client.ApiClient],
    namespace: str,
    job_name: str,
    backoff_interval: float = 1,
    maximum_backoff: float = 32,
    exponential_backoff: bool = False,
    fail_on_container_waiting_reasons: Optional[List[str]] = None,
    stream_logs: bool = True,
    container_name: Optional[str] = None,
    api_request_timeout: Optional[int] = None,
) -> None:
    """Wait for a job to finish.

    Args:
        get_client: A function that returns a Kubernetes API client.
        namespace: Kubernetes namespace.
        job_name: Name of the job for which to wait.
        backoff_interval: The interval to wait between polling the job status.
        maximum_backoff: The maximum interval to wait between polling the job
            status.
        exponential_backoff: Whether to use exponential backoff.
        fail_on_container_waiting_reasons: List of container waiting reasons
            that will cause the job to fail.
        stream_logs: Whether to stream the job logs.
        container_name: Name of the container to stream logs from.
        api_request_timeout: The request timeout in seconds.

    Raises:
        RuntimeError: If the job failed or timed out.
    """
    logged_lines_per_pod: Dict[str, int] = defaultdict(int)
    finished_pods = set()

    while True:
        client = get_client()
        batch_api = k8s_client.BatchV1Api(client)
        core_api = k8s_client.CoreV1Api(client)

        job: k8s_client.V1Job = retry_on_api_exception(
            batch_api.read_namespaced_job,
            api_request_timeout=api_request_timeout,
        )(name=job_name, namespace=namespace)

        if job.status.conditions:
            for condition in job.status.conditions:
                if condition.type == "Complete" and condition.status == "True":
                    return
                if condition.type == "Failed" and condition.status == "True":
                    raise RuntimeError(
                        f"Job `{namespace}:{job_name}` failed: "
                        f"{condition.message}"
                    )

        if fail_on_container_waiting_reasons:
            pod_list: k8s_client.V1PodList = retry_on_api_exception(
                core_api.list_namespaced_pod,
                api_request_timeout=api_request_timeout,
            )(
                namespace=namespace,
                label_selector=f"job-name={job_name}",
                field_selector="status.phase=Pending",
            )
            for pod in pod_list.items:
                container_state = get_container_status(
                    pod, container_name or "main"
                )

                if (
                    container_state
                    and (waiting_state := container_state.waiting)
                    and waiting_state.reason
                    in fail_on_container_waiting_reasons
                ):
                    retry_on_api_exception(
                        batch_api.delete_namespaced_job,
                        api_request_timeout=api_request_timeout,
                    )(
                        name=job_name,
                        namespace=namespace,
                        propagation_policy="Foreground",
                    )
                    raise RuntimeError(
                        f"Job `{namespace}:{job_name}` failed: "
                        f"Detected container in state "
                        f"{waiting_state.reason}"
                    )

        if stream_logs:
            try:
                pod_list = core_api.list_namespaced_pod(
                    namespace=namespace,
                    label_selector=f"job-name={job_name}",
                )
            except ApiException as e:
                logger.error("Error fetching pods: %s.", e)
                pod_list = []
            else:
                # Sort pods by creation timestamp, oldest first
                pod_list.items.sort(
                    key=lambda pod: pod.metadata.creation_timestamp,
                )

            for pod in pod_list.items:
                pod_name = pod.metadata.name
                pod_status = pod.status.phase

                if pod_name in finished_pods:
                    # We've already streamed all logs for this pod, so we can
                    # skip it.
                    continue

                if pod_status == PodPhase.PENDING.value:
                    # The pod is still pending, so we can't stream logs for it
                    # yet.
                    continue

                if pod_status in [
                    PodPhase.SUCCEEDED.value,
                    PodPhase.FAILED.value,
                ]:
                    finished_pods.add(pod_name)

                containers = pod.spec.containers
                if not container_name:
                    container_name = containers[0].name

                try:
                    response = core_api.read_namespaced_pod_log(
                        name=pod_name,
                        namespace=namespace,
                        container=container_name,
                        _preload_content=False,
                    )
                except ApiException as e:
                    logger.error("Error reading pod logs: %s.", e)
                else:
                    raw_data = response.data
                    decoded_log = raw_data.decode("utf-8", errors="replace")
                    logs = decoded_log.splitlines()
                    logged_lines = logged_lines_per_pod[pod_name]
                    if len(logs) > logged_lines:
                        for line in logs[logged_lines:]:
                            logger.info(line)
                        logged_lines_per_pod[pod_name] = len(logs)

        time.sleep(backoff_interval)
        if exponential_backoff and backoff_interval < maximum_backoff:
            backoff_interval *= 2


def check_job_status(
    batch_api: k8s_client.BatchV1Api,
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    job_name: str,
    fail_on_container_waiting_reasons: Optional[List[str]] = None,
    container_name: Optional[str] = None,
    api_request_timeout: Optional[int] = None,
) -> Tuple[JobStatus, Optional[str]]:
    """Check the status of a job.

    Args:
        batch_api: Kubernetes BatchV1Api client.
        core_api: Kubernetes CoreV1Api client.
        namespace: Kubernetes namespace.
        job_name: Name of the job for which to wait.
        fail_on_container_waiting_reasons: List of container waiting reasons
            that will cause the job to fail.
        container_name: Name of the container to check for failure.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The status of the job and an error message if the job failed.
    """
    job: k8s_client.V1Job = retry_on_api_exception(
        batch_api.read_namespaced_job, api_request_timeout=api_request_timeout
    )(name=job_name, namespace=namespace)

    if job.status.conditions:
        for condition in job.status.conditions:
            if condition.type == "Complete" and condition.status == "True":
                return JobStatus.SUCCEEDED, None
            if condition.type == "Failed" and condition.status == "True":
                error_message = condition.message or "Unknown"
                container_failure_reason = None
                try:
                    pods = core_api.list_namespaced_pod(
                        label_selector=f"job-name={job_name}",
                        namespace=namespace,
                    ).items
                    # Sort pods by creation timestamp, oldest first
                    pods.sort(
                        key=lambda pod: pod.metadata.creation_timestamp,
                    )
                    if pods:
                        if (
                            termination_reason
                            := get_container_termination_reason(
                                pods[-1], container_name or "main"
                            )
                        ):
                            exit_code, reason = termination_reason
                            if exit_code != 0:
                                container_failure_reason = (
                                    f"{reason}, exit_code={exit_code}"
                                )
                except Exception:
                    pass

                if container_failure_reason:
                    error_message += f" (container failure reason: {container_failure_reason})"

                return JobStatus.FAILED, error_message

    if fail_on_container_waiting_reasons:
        pod_list: k8s_client.V1PodList = retry_on_api_exception(
            core_api.list_namespaced_pod,
            api_request_timeout=api_request_timeout,
        )(
            namespace=namespace,
            label_selector=f"job-name={job_name}",
            field_selector="status.phase=Pending",
        )
        for pod in pod_list.items:
            container_state = get_container_status(
                pod, container_name or "main"
            )

            if (
                container_state
                and (waiting_state := container_state.waiting)
                and waiting_state.reason in fail_on_container_waiting_reasons
            ):
                retry_on_api_exception(
                    batch_api.delete_namespaced_job,
                    api_request_timeout=api_request_timeout,
                )(
                    name=job_name,
                    namespace=namespace,
                    propagation_policy="Foreground",
                )
                error_message = (
                    f"Detected container in state `{waiting_state.reason}`"
                )
                return JobStatus.FAILED, error_message

    return JobStatus.RUNNING, None


def create_config_map(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    name: str,
    data: Dict[str, str],
    api_request_timeout: Optional[int] = None,
) -> None:
    """Create a Kubernetes config map.

    Args:
        core_api: Kubernetes CoreV1Api client.
        namespace: Kubernetes namespace.
        name: Name of the config map to create.
        data: Data to store in the config map.
        api_request_timeout: The request timeout in seconds.
    """
    retry_on_api_exception(
        core_api.create_namespaced_config_map,
        api_request_timeout=api_request_timeout,
    )(
        namespace=namespace,
        body=k8s_client.V1ConfigMap(metadata={"name": name}, data=data),
    )


def update_config_map(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    name: str,
    data: Dict[str, str],
    api_request_timeout: Optional[int] = None,
) -> None:
    """Update a Kubernetes config map.

    Args:
        core_api: Kubernetes CoreV1Api client.
        namespace: Kubernetes namespace.
        name: Name of the config map to update.
        data: Data to store in the config map.
        api_request_timeout: The request timeout in seconds.
    """
    retry_on_api_exception(
        core_api.patch_namespaced_config_map,
        api_request_timeout=api_request_timeout,
    )(
        namespace=namespace,
        name=name,
        body=k8s_client.V1ConfigMap(data=data),
    )


def get_config_map(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    name: str,
    api_request_timeout: Optional[int] = None,
) -> k8s_client.V1ConfigMap:
    """Get a Kubernetes config map.

    Args:
        core_api: Kubernetes CoreV1Api client.
        namespace: Kubernetes namespace.
        name: Name of the config map to get.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The config map.
    """
    return retry_on_api_exception(
        core_api.read_namespaced_config_map,
        api_request_timeout=api_request_timeout,
    )(
        namespace=namespace,
        name=name,
    )


def delete_config_map(
    core_api: k8s_client.CoreV1Api,
    namespace: str,
    name: str,
    api_request_timeout: Optional[int] = None,
) -> None:
    """Delete a Kubernetes config map.

    Args:
        core_api: Kubernetes CoreV1Api client.
        namespace: Kubernetes namespace.
        name: Name of the config map to delete.
        api_request_timeout: The request timeout in seconds.
    """
    retry_on_api_exception(
        core_api.delete_namespaced_config_map,
        api_request_timeout=api_request_timeout,
    )(
        namespace=namespace,
        name=name,
    )


def get_parent_job_name(
    core_api: k8s_client.CoreV1Api,
    pod_name: str,
    namespace: str,
    api_request_timeout: Optional[int] = None,
) -> Optional[str]:
    """Get the name of the job that created a pod.

    Args:
        core_api: Kubernetes CoreV1Api client.
        pod_name: Name of the pod.
        namespace: Kubernetes namespace.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The name of the job that created the pod, or None if the pod is not
        associated with a job.
    """
    pod = get_pod(
        core_api,
        pod_name=pod_name,
        namespace=namespace,
        api_request_timeout=api_request_timeout,
    )
    if (
        pod
        and pod.metadata
        and pod.metadata.labels
        and (job_name := pod.metadata.labels.get("job-name", None))
    ):
        return cast(str, job_name)

    return None


def apply_default_resource_requests(
    memory: str,
    cpu: Optional[str] = None,
    pod_settings: Optional[KubernetesPodSettings] = None,
) -> KubernetesPodSettings:
    """Applies default resource requests to a pod settings object.

    Args:
        memory: The memory resource request.
        cpu: The CPU resource request.
        pod_settings: The pod settings to update. A new one will be created
            if not provided.

    Returns:
        The new or updated pod settings.
    """
    resources = {
        "requests": {"memory": memory},
    }
    if cpu:
        resources["requests"]["cpu"] = cpu
    if not pod_settings:
        pod_settings = KubernetesPodSettings(resources=resources)
    elif not pod_settings.resources:
        pod_settings = pod_settings.model_copy(update={"resources": resources})
    else:
        set_requests = pod_settings.resources.get("requests", {})
        resources["requests"].update(set_requests)
        pod_settings.resources["requests"] = resources["requests"]

    return pod_settings


# ============================================================================
# Resource Conversion Utilities
# ============================================================================


def convert_resource_settings_to_k8s_format(
    resource_settings: "ResourceSettings",
) -> Tuple[Dict[str, str], Dict[str, str], int]:
    """Convert ZenML ResourceSettings to Kubernetes resource format.

    Args:
        resource_settings: The resource settings from pipeline configuration.

    Returns:
        Tuple of (requests, limits, replicas) in Kubernetes format.
        - requests: Dict with 'cpu', 'memory', and optionally 'nvidia.com/gpu' keys
        - limits: Dict with same keys as requests (GPUs require limits=requests)
        - replicas: Number of replicas
    """
    requests: Dict[str, str] = {}
    limits: Dict[str, str] = {}

    if resource_settings.cpu_count is not None:
        cpu_value = resource_settings.cpu_count
        if cpu_value == int(cpu_value):
            cpu_str = str(int(cpu_value))
        else:
            cpu_str = f"{int(cpu_value * 1000)}m"

        requests["cpu"] = cpu_str

    if memory_mib := resource_settings.get_memory(unit=ByteUnit.MIB):
        memory_str = f"{int(memory_mib)}Mi"
        requests["memory"] = memory_str

    min_r = resource_settings.min_replicas
    max_r = resource_settings.max_replicas

    if min_r is not None and min_r > 0:
        replicas = min_r
    elif max_r is not None and max_r > 0:
        replicas = max_r
    else:
        replicas = 1

    if resource_settings.gpu_count:
        gpu_str = str(resource_settings.gpu_count)
        requests["nvidia.com/gpu"] = gpu_str
        limits["nvidia.com/gpu"] = gpu_str
        logger.info(
            f"Configured {resource_settings.gpu_count} GPU(s) per pod. "
            f"Ensure your cluster has GPU nodes with the nvidia.com/gpu resource. "
            f"You may need to install the NVIDIA device plugin: "
            f"https://github.com/NVIDIA/k8s-device-plugin"
        )

    return requests, limits, replicas


# ============================================================================
# URL Building Utilities
# ============================================================================


def build_service_url(
    core_api: k8s_client.CoreV1Api,
    service: Union[k8s_client.V1Service, Dict[str, Any]],
    namespace: str,
    ingress: Optional[Union[k8s_client.V1Ingress, Dict[str, Any]]] = None,
) -> Optional[str]:
    """Build URL for accessing a Kubernetes service.

    Supports multiple service types:
    - LoadBalancer: Returns http://<external-ip>:<port> if IP is assigned
    - NodePort: Returns http://<node-ip>:<node-port> if nodes have external IPs
    - ClusterIP: Returns internal cluster URL with warning about port-forwarding

    Args:
        core_api: Kubernetes Core V1 API client.
        service: Kubernetes Service resource (model or dict).
        namespace: Kubernetes namespace.
        ingress: Optional Kubernetes Ingress resource.

    Returns:
        Service URL
    """
    service_dict = normalize_resource_to_dict(service)

    if ingress:
        ingress_dict = normalize_resource_to_dict(ingress)

        ingress_spec = ingress_dict.get("spec")
        if ingress_spec:
            protocol = "https" if ingress_spec.get("tls") else "http"
            rules = ingress_spec.get("rules", [])
            if rules:
                rule = rules[0]
                host = rule.get("host")
                http_config = rule.get("http", {})
                paths = http_config.get("paths", [])
                if host and paths:
                    path = paths[0].get("path", "/")
                    return f"{protocol}://{host}{path}"

    metadata = service_dict.get("metadata", {})
    spec = service_dict.get("spec", {})

    service_name = metadata.get("name", "unknown")
    service_type = spec.get("type")
    ports = spec.get("ports", [])

    if not service_type or not ports:
        return None

    service_port = ports[0].get("port")

    if service_type == "LoadBalancer":
        status = service_dict.get("status", {})
        lb_status = status.get("loadBalancer", {})
        ingress_list = lb_status.get("ingress", [])

        if ingress_list:
            lb_ingress = ingress_list[0]
            host = lb_ingress.get("ip") or lb_ingress.get("hostname")
            if host:
                return f"http://{host}:{service_port}"

        return None

    if service_type == "NodePort":
        node_port = ports[0].get("nodePort")
        if not node_port:
            return None

        try:
            nodes = core_api.list_node()

            for node in nodes.items:
                if node.status and node.status.addresses:
                    for address in node.status.addresses:
                        if address.type == "ExternalIP":
                            return f"http://{address.address}:{node_port}"

            for node in nodes.items:
                if node.status and node.status.addresses:
                    for address in node.status.addresses:
                        if address.type == "InternalIP":
                            logger.warning(
                                f"NodePort service '{service_name}' has no ExternalIP. "
                                f"Use: kubectl port-forward -n {namespace} "
                                f"service/{service_name} 8080:{service_port}"
                            )
                            return f"http://{address.address}:{node_port}"
        except Exception as e:
            logger.error(f"Failed to get node IPs: {e}")

        return None

    if service_type == "ClusterIP":
        logger.warning(
            f"Service '{service_name}' uses ClusterIP (cluster-internal only). "
            f"Use: kubectl port-forward -n {namespace} "
            f"service/{service_name} 8080:{service_port}"
        )
        return f"http://{service_name}.{namespace}.svc.cluster.local:{service_port}"

    return None



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline snapshot models."""

from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from pydantic import Field, field_validator

from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.pipeline_run_configuration import PipelineRunConfiguration
from zenml.config.pipeline_spec import PipelineSpec
from zenml.config.step_configurations import Step
from zenml.constants import STR_FIELD_MAX_LENGTH, TEXT_FIELD_MAX_LENGTH
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.models.v2.base.base import BaseUpdate, BaseZenModel
from zenml.models.v2.base.scoped import (
    ProjectScopedFilter,
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
    TaggableFilter,
)
from zenml.models.v2.core.code_reference import (
    CodeReferenceRequest,
    CodeReferenceResponse,
)
from zenml.models.v2.core.deployment import DeploymentResponse
from zenml.models.v2.core.pipeline import PipelineResponse
from zenml.models.v2.core.pipeline_build import (
    PipelineBuildResponse,
)
from zenml.models.v2.core.schedule import ScheduleResponse
from zenml.models.v2.core.stack import StackResponse
from zenml.models.v2.core.tag import TagResponse
from zenml.models.v2.core.user import UserResponse

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.models.v2.core.curated_visualization import (
        CuratedVisualizationResponse,
    )
    from zenml.zen_stores.schemas.base_schemas import BaseSchema

    AnySchema = TypeVar("AnySchema", bound=BaseSchema)
    AnyQuery = TypeVar("AnyQuery", bound=Any)


# ------------------ Request Model ------------------


class PipelineSnapshotBase(BaseZenModel):
    """Base model for pipeline snapshots."""

    run_name_template: str = Field(
        title="The run name template for runs created using this snapshot.",
    )
    pipeline_configuration: PipelineConfiguration = Field(
        title="The pipeline configuration for this snapshot."
    )
    step_configurations: Dict[str, Step] = Field(
        default={}, title="The step configurations for this snapshot."
    )
    client_environment: Dict[str, Any] = Field(
        default={}, title="The client environment for this snapshot."
    )
    client_version: Optional[str] = Field(
        default=None,
        title="The version of the ZenML installation on the client side.",
    )
    server_version: Optional[str] = Field(
        default=None,
        title="The version of the ZenML installation on the server side.",
    )
    pipeline_version_hash: Optional[str] = Field(
        default=None,
        title="The pipeline version hash of the snapshot.",
    )
    pipeline_spec: Optional[PipelineSpec] = Field(
        default=None,
        title="The pipeline spec of the snapshot.",
    )
    is_dynamic: bool = Field(
        default=False,
        title="Whether this is a snapshot of a dynamic pipeline.",
    )


class PipelineSnapshotRequest(PipelineSnapshotBase, ProjectScopedRequest):
    """Request model for pipeline snapshots."""

    name: Optional[Union[str, bool]] = Field(
        default=None,
        title="The name of the snapshot.",
    )
    description: Optional[str] = Field(
        default=None,
        title="The description of the snapshot.",
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    source_code: Optional[str] = Field(
        default=None,
        title="The source code of the pipeline function or class.",
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    replace: Optional[bool] = Field(
        default=None,
        title="Whether to replace the existing snapshot with the same name.",
    )
    tags: Optional[List[str]] = Field(
        default=None,
        title="Tags of the snapshot.",
    )

    stack: UUID = Field(title="The stack associated with the snapshot.")
    pipeline: UUID = Field(title="The pipeline associated with the snapshot.")
    build: Optional[UUID] = Field(
        default=None, title="The build associated with the snapshot."
    )
    schedule: Optional[UUID] = Field(
        default=None, title="The schedule associated with the snapshot."
    )
    code_reference: Optional["CodeReferenceRequest"] = Field(
        default=None,
        title="The code reference associated with the snapshot.",
    )
    code_path: Optional[str] = Field(
        default=None,
        title="Optional path where the code is stored in the artifact store.",
    )
    template: Optional[UUID] = Field(
        default=None,
        description="DEPRECATED: Template used for the snapshot.",
    )
    source_snapshot: Optional[UUID] = Field(
        default=None,
        description="Snapshot that is the source of this snapshot.",
    )

    @field_validator("name")
    @classmethod
    def _validate_name(cls, v: Any) -> Any:
        if isinstance(v, str):
            if not v:
                raise ValueError("Snapshot name cannot be empty.")
            if len(v) > STR_FIELD_MAX_LENGTH:
                raise ValueError(
                    f"Snapshot name `{v}` is too long. The maximum length "
                    f"is {STR_FIELD_MAX_LENGTH} characters."
                )
        elif v is True:
            raise ValueError("Snapshot name cannot be `True`.")

        return v


# ------------------ Update Model ------------------


class PipelineSnapshotUpdate(BaseUpdate):
    """Pipeline snapshot update model."""

    name: Optional[Union[str, bool]] = Field(
        default=None,
        title="The name of the snapshot. If set to "
        "False, the name will be removed.",
    )
    description: Optional[str] = Field(
        default=None,
        title="The description of the snapshot.",
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    replace: Optional[bool] = Field(
        default=None,
        title="Whether to replace the existing snapshot with the same name.",
    )
    add_tags: Optional[List[str]] = Field(
        default=None, title="New tags to add to the snapshot."
    )
    remove_tags: Optional[List[str]] = Field(
        default=None, title="Tags to remove from the snapshot."
    )

    @field_validator("name")
    @classmethod
    def _validate_name(cls, v: Any) -> Any:
        if isinstance(v, str):
            if not v:
                raise ValueError("Snapshot name cannot be empty.")
            if len(v) > STR_FIELD_MAX_LENGTH:
                raise ValueError(
                    f"Snapshot name `{v}` is too long. The maximum length "
                    f"is {STR_FIELD_MAX_LENGTH} characters."
                )
        elif v is True:
            raise ValueError("Snapshot name cannot be `True`.")

        return v


# ------------------ Response Model ------------------


class PipelineSnapshotResponseBody(ProjectScopedResponseBody):
    """Response body for pipeline snapshots."""

    runnable: bool = Field(
        title="If a run can be started from the snapshot.",
    )
    deployable: bool = Field(
        title="If the snapshot can be deployed.",
    )
    is_dynamic: bool = Field(
        title="Whether this is a snapshot of a dynamic pipeline.",
    )


class PipelineSnapshotResponseMetadata(ProjectScopedResponseMetadata):
    """Response metadata for pipeline snapshots."""

    __zenml_skip_dehydration__: ClassVar[List[str]] = [
        "pipeline_configuration",
        "step_configurations",
        "client_environment",
        "pipeline_spec",
    ]

    description: Optional[str] = Field(
        default=None,
        title="The description of the snapshot.",
    )
    source_code: Optional[str] = Field(
        default=None,
        title="The source code of the pipeline function or class.",
        max_length=TEXT_FIELD_MAX_LENGTH,
    )
    run_name_template: str = Field(
        title="The run name template for runs created using this snapshot.",
    )
    pipeline_configuration: PipelineConfiguration = Field(
        title="The pipeline configuration for this snapshot."
    )
    step_configurations: Dict[str, Step] = Field(
        default={}, title="The step configurations for this snapshot."
    )
    client_environment: Dict[str, Any] = Field(
        default={}, title="The client environment for this snapshot."
    )
    client_version: Optional[str] = Field(
        title="The version of the ZenML installation on the client side."
    )
    server_version: Optional[str] = Field(
        title="The version of the ZenML installation on the server side."
    )
    pipeline_version_hash: Optional[str] = Field(
        default=None, title="The pipeline version hash of the snapshot."
    )
    pipeline_spec: Optional[PipelineSpec] = Field(
        default=None, title="The pipeline spec of the snapshot."
    )
    code_path: Optional[str] = Field(
        default=None,
        title="Optional path where the code is stored in the artifact store.",
    )
    template_id: Optional[UUID] = Field(
        default=None,
        description="Template from which this snapshot was created.",
        deprecated=True,
    )
    source_snapshot_id: Optional[UUID] = Field(
        default=None,
        description="Snapshot that is the source of this snapshot.",
    )
    config_template: Optional[Dict[str, Any]] = Field(
        default=None, title="Run configuration template."
    )
    config_schema: Optional[Dict[str, Any]] = Field(
        default=None, title="Run configuration schema."
    )


class PipelineSnapshotResponseResources(ProjectScopedResponseResources):
    """Run snapshot resources."""

    pipeline: PipelineResponse = Field(
        title="The pipeline associated with the snapshot."
    )
    stack: Optional[StackResponse] = Field(
        default=None, title="The stack associated with the snapshot."
    )
    build: Optional[PipelineBuildResponse] = Field(
        default=None,
        title="The pipeline build associated with the snapshot.",
    )
    schedule: Optional[ScheduleResponse] = Field(
        default=None, title="The schedule associated with the snapshot."
    )
    code_reference: Optional[CodeReferenceResponse] = Field(
        default=None,
        title="The code reference associated with the snapshot.",
    )
    deployment: Optional[DeploymentResponse] = Field(
        default=None,
        title="The deployment associated with the snapshot.",
    )
    tags: List[TagResponse] = Field(
        default=[],
        title="Tags associated with the snapshot.",
    )
    latest_run_id: Optional[UUID] = Field(
        default=None,
        title="The ID of the latest run of the snapshot.",
    )
    latest_run_status: Optional[ExecutionStatus] = Field(
        default=None,
        title="The status of the latest run of the snapshot.",
    )
    latest_run_user: Optional[UserResponse] = Field(
        default=None,
        title="The user that created the latest run of the snapshot.",
    )
    visualizations: List["CuratedVisualizationResponse"] = Field(
        default=[],
        title="Curated visualizations associated with the pipeline snapshot.",
    )


class PipelineSnapshotResponse(
    ProjectScopedResponse[
        PipelineSnapshotResponseBody,
        PipelineSnapshotResponseMetadata,
        PipelineSnapshotResponseResources,
    ]
):
    """Response model for pipeline snapshots."""

    name: Optional[str] = Field(
        default=None,
        title="The name of the snapshot.",
        max_length=STR_FIELD_MAX_LENGTH,
    )

    def get_hydrated_version(self) -> "PipelineSnapshotResponse":
        """Return the hydrated version of this pipeline snapshot.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_snapshot(self.id)

    # Body and metadata properties

    @property
    def runnable(self) -> bool:
        """The `runnable` property.

        Returns:
            the value of the property.
        """
        return self.get_body().runnable

    @property
    def deployable(self) -> bool:
        """The `deployable` property.

        Returns:
            the value of the property.
        """
        return self.get_body().deployable

    @property
    def is_dynamic(self) -> bool:
        """The `is_dynamic` property.

        Returns:
            the value of the property.
        """
        return self.get_body().is_dynamic

    @property
    def description(self) -> Optional[str]:
        """The `description` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().description

    @property
    def source_code(self) -> Optional[str]:
        """The `source_code` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().source_code

    @property
    def run_name_template(self) -> str:
        """The `run_name_template` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().run_name_template

    @property
    def pipeline_configuration(self) -> PipelineConfiguration:
        """The `pipeline_configuration` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().pipeline_configuration

    @property
    def step_configurations(self) -> Dict[str, Step]:
        """The `step_configurations` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().step_configurations

    @property
    def client_environment(self) -> Dict[str, Any]:
        """The `client_environment` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().client_environment

    @property
    def client_version(self) -> Optional[str]:
        """The `client_version` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().client_version

    @property
    def server_version(self) -> Optional[str]:
        """The `server_version` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().server_version

    @property
    def pipeline_version_hash(self) -> Optional[str]:
        """The `pipeline_version_hash` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().pipeline_version_hash

    @property
    def pipeline_spec(self) -> Optional[PipelineSpec]:
        """The `pipeline_spec` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().pipeline_spec

    @property
    def code_path(self) -> Optional[str]:
        """The `code_path` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().code_path

    @property
    def template_id(self) -> Optional[UUID]:
        """The `template_id` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().template_id

    @property
    def source_snapshot_id(self) -> Optional[UUID]:
        """The `source_snapshot_id` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().source_snapshot_id

    @property
    def config_schema(self) -> Optional[Dict[str, Any]]:
        """The `config_schema` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().config_schema

    @property
    def config_template(self) -> Optional[Dict[str, Any]]:
        """The `config_template` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().config_template

    @property
    def pipeline(self) -> PipelineResponse:
        """The `pipeline` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().pipeline

    @property
    def stack(self) -> Optional[StackResponse]:
        """The `stack` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().stack

    @property
    def build(self) -> Optional[PipelineBuildResponse]:
        """The `build` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().build

    @property
    def schedule(self) -> Optional[ScheduleResponse]:
        """The `schedule` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().schedule

    @property
    def code_reference(self) -> Optional[CodeReferenceResponse]:
        """The `code_reference` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().code_reference

    @property
    def deployment(self) -> Optional[DeploymentResponse]:
        """The `deployment` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().deployment

    @property
    def tags(self) -> List[TagResponse]:
        """The `tags` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().tags

    @property
    def latest_run_id(self) -> Optional[UUID]:
        """The `latest_run_id` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().latest_run_id

    @property
    def latest_run_status(self) -> Optional[ExecutionStatus]:
        """The `latest_run_status` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().latest_run_status

    @property
    def latest_run_user(self) -> Optional[UserResponse]:
        """The `latest_run_user` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().latest_run_user


# ------------------ Filter Model ------------------


class PipelineSnapshotFilter(ProjectScopedFilter, TaggableFilter):
    """Model for filtering pipeline snapshots."""

    FILTER_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *ProjectScopedFilter.FILTER_EXCLUDE_FIELDS,
        *TaggableFilter.FILTER_EXCLUDE_FIELDS,
        "named_only",
        "pipeline",
        "stack",
        "runnable",
        "deployable",
        "deployed",
        "trigger_id",
    ]
    CUSTOM_SORTING_OPTIONS = [
        *ProjectScopedFilter.CUSTOM_SORTING_OPTIONS,
        *TaggableFilter.CUSTOM_SORTING_OPTIONS,
        "pipeline",
        "stack",
        "deployment",
    ]
    CLI_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *ProjectScopedFilter.CLI_EXCLUDE_FIELDS,
        *TaggableFilter.CLI_EXCLUDE_FIELDS,
    ]

    name: Optional[str] = Field(
        default=None,
        description="Name of the snapshot.",
    )
    named_only: Optional[bool] = Field(
        default=None,
        description="Whether to only return snapshots with a name.",
    )
    pipeline: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Pipeline associated with the snapshot.",
        union_mode="left_to_right",
    )
    stack: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Stack associated with the snapshot.",
        union_mode="left_to_right",
    )
    build_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Build associated with the snapshot.",
        union_mode="left_to_right",
    )
    schedule_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Schedule associated with the snapshot.",
        union_mode="left_to_right",
    )
    source_snapshot_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Source snapshot used for the snapshot.",
        union_mode="left_to_right",
    )
    runnable: Optional[bool] = Field(
        default=None,
        description="Whether the snapshot is runnable.",
    )
    deployable: Optional[bool] = Field(
        default=None,
        description="Whether the snapshot is deployable.",
    )
    deployed: Optional[bool] = Field(
        default=None,
        description="Whether the snapshot is deployed.",
    )
    trigger_id: UUID | None = Field(
        default=None,
        description="Trigger associated with the snapshot (attached).",
    )

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        from sqlmodel import and_, col, not_, select

        from zenml.zen_stores.schemas import (
            DeploymentSchema,
            PipelineBuildSchema,
            PipelineSchema,
            PipelineSnapshotSchema,
            StackComponentSchema,
            StackCompositionSchema,
            StackSchema,
        )

        custom_filters = super().get_custom_filters(table)

        if self.named_only:
            custom_filters.append(
                col(PipelineSnapshotSchema.name).is_not(None)
            )

        if self.pipeline:
            pipeline_filter = and_(
                PipelineSnapshotSchema.pipeline_id == PipelineSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.pipeline, table=PipelineSchema
                ),
            )
            custom_filters.append(pipeline_filter)

        if self.stack:
            stack_filter = and_(
                PipelineSnapshotSchema.stack_id == StackSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.stack,
                    table=StackSchema,
                ),
            )
            custom_filters.append(stack_filter)

        if self.runnable is True:
            runnable_filter = and_(
                # The following condition is not perfect as it does not
                # consider stacks with custom flavor components or local
                # components, but the best we can do currently with our
                # table columns.
                PipelineSnapshotSchema.build_id == PipelineBuildSchema.id,
                col(PipelineBuildSchema.is_local).is_(False),
                col(PipelineBuildSchema.stack_id).is_not(None),
            )

            custom_filters.append(runnable_filter)

        if self.deployable is True:
            deployer_exists = (
                select(StackComponentSchema.id)
                .where(
                    StackComponentSchema.type
                    == StackComponentType.DEPLOYER.value
                )
                .where(
                    StackCompositionSchema.component_id
                    == StackComponentSchema.id
                )
                .where(
                    StackCompositionSchema.stack_id
                    == PipelineSnapshotSchema.stack_id
                )
                .exists()
            )
            deployable_filter = and_(
                col(PipelineSnapshotSchema.build_id).is_not(None),
                deployer_exists,
            )

            custom_filters.append(deployable_filter)

        if self.deployed is not None:
            deployment_exists = (
                select(DeploymentSchema.id)
                .where(
                    DeploymentSchema.snapshot_id == PipelineSnapshotSchema.id
                )
                .exists()
            )
            if self.deployed is True:
                deployed_filter = and_(deployment_exists)
            else:
                deployed_filter = and_(not_(deployment_exists))
            custom_filters.append(deployed_filter)

        return custom_filters

    def apply_sorting(
        self,
        query: "AnyQuery",
        table: Type["AnySchema"],
    ) -> "AnyQuery":
        """Apply sorting to the query.

        Args:
            query: The query to which to apply the sorting.
            table: The query table.

        Returns:
            The query with sorting applied.
        """
        from sqlmodel import asc, desc

        from zenml.enums import SorterOps
        from zenml.zen_stores.schemas import (
            DeploymentSchema,
            PipelineSchema,
            PipelineSnapshotSchema,
            StackSchema,
        )

        sort_by, operand = self.sorting_params

        if sort_by == "pipeline":
            query = query.outerjoin(
                PipelineSchema,
                PipelineSnapshotSchema.pipeline_id == PipelineSchema.id,
            )
            column = PipelineSchema.name
        elif sort_by == "stack":
            query = query.outerjoin(
                StackSchema,
                PipelineSnapshotSchema.stack_id == StackSchema.id,
            )
            column = StackSchema.name
        elif sort_by == "deployment":
            query = query.outerjoin(
                DeploymentSchema,
                PipelineSnapshotSchema.id == DeploymentSchema.snapshot_id,
            )
            column = DeploymentSchema.name
        else:
            return super().apply_sorting(query=query, table=table)

        query = query.add_columns(column)

        if operand == SorterOps.ASCENDING:
            query = query.order_by(asc(column))
        else:
            query = query.order_by(desc(column))

        return query

    def apply_filter(
        self, query: "AnyQuery", table: Type["AnySchema"]
    ) -> "AnyQuery":
        """Applies the filter to a query.

        Args:
            query: The query to which to apply the filter.
            table: The query table.

        Returns:
            The query with filter applied.
        """
        from zenml.zen_stores.schemas import (
            PipelineSnapshotSchema,
            TriggerSnapshotSchema,
        )

        query = super().apply_filter(query=query, table=table)

        if self.trigger_id is not None:
            query = query.join(
                TriggerSnapshotSchema,
                PipelineSnapshotSchema.id == TriggerSnapshotSchema.snapshot_id,
            ).where(TriggerSnapshotSchema.trigger_id == self.trigger_id)

        return query


class PipelineSnapshotRunRequest(BaseZenModel):
    """Request model for running a pipeline snapshot."""

    run_configuration: Optional[PipelineRunConfiguration] = Field(
        default=None,
        title="The run configuration for the snapshot.",
    )
    step_run: Optional[UUID] = Field(
        default=None,
        title="The ID of the step run that ran the snapshot.",
    )



"""add api transaction result table [73c397e3c3ba].

Revision ID: 73c397e3c3ba
Revises: 0.93.2
Create Date: 2026-01-23 18:51:29.944099

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.dialects.mysql import MEDIUMBLOB

# revision identifiers, used by Alembic.
revision = "73c397e3c3ba"
down_revision = "0.93.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    op.create_table(
        "api_transaction_result",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column(
            "result",
            MEDIUMBLOB
            if op.get_bind().dialect.name == "mysql"
            else sa.LargeBinary(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["id"],
            ["api_transaction.id"],
            name="fk_api_transaction_result_id_api_transaction",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("api_transaction", schema=None) as batch_op:
        batch_op.drop_column("result")


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("api_transaction", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("result", sa.VARCHAR(length=16777215), nullable=True)
        )

    op.drop_table("api_transaction_result")
